
from turtle import color
import numpy as np
import scipy
import matplotlib.pyplot as plt
from matplotlib.patches import ConnectionPatch
from matplotlib.widgets import Button, TextBox
import matplotlib.patheffects as PathEffects

import plotly.express as px
import plotly.graph_objects as go
from ipywidgets import widgets, interactive, HBox, VBox, Output, Layout

import pandas as pd
from scipy.stats import binom
from sklearn.decomposition import PCA, FastICA

from sklearn.neighbors import NearestNeighbors

from umap import UMAP


class SpatialGraph():
    """SpatialGraph: Container object for all KNN-graph based operations on spatial data sets.

    :param sdata: The spatial data source 
    :type sdata: plankton.SpatialData
    :param n_neighbors: number of nearset neighbors to infer, defaults to 10
    :type n_neighbors: int, optional
    """

    def __init__(self, sdata, n_neighbors=10) -> None:

        self.sdata = sdata
        self.n_neighbors = n_neighbors
        self._neighbors = None
        self._neighbor_types = None
        self._distances = None
        self._umap = None
        self._tsne = None

    @property
    def neighbors(self):
        """neighbors: Returns an array of neighbors, where each entry encoding the row position of a point in the data set.

        :return: 2d array of n-neighbors per sample data point
        :rtype: numpy.Array
        """
        if self._neighbors is None:
            self._distances, self._neighbors, self._neighbor_types = self.update_knn(
                self.n_neighbors)
        return self._neighbors[:, :self.n_neighbors]

    @property
    def distances(self):
        """distances: Returns an array of inter-spot distances, which each entry encoding the distance to a defined neighbor. 

        :return: 2d array of distances to the data points of self.neighbors.
        :rtype: numpy.Array
        """
        if self._distances is None:
            self._distances, self._neighbors, self._neighbor_types = self.update_knn(
                self.n_neighbors)
        return self._distances[:, :self.n_neighbors]

    @property
    def neighbor_types(self):
        """neighbor_types: Returns an array of gene_ids for the spots listed in self.neighbors

        :return: 2d array of gene_ids.
        :rtype: numpy.Array
        """
        if self._neighbor_types is None:
            self._distances, self._neighbors, self._neighbor_types = self.update_knn(
                self.n_neighbors)
        return self._neighbor_types[:, :self.n_neighbors]

    @property
    def umap_0(self):
        return self._umap[:, 0]

    @property
    def umap_1(self):
        return self._umap[:, 1]

    @property
    def umap(self):
        """umap: Returns the coordinates of the UMAP representation of the source data.

        :return: UMAP coordinates
        :rtype: numpy.Array
        """
        if self._umap is None:
            print("No UMAP calculated at this point. Please call 'run_umap' first.")
        return self._umap

    def __getitem__(self, *args):
        sg = SpatialGraph(self.sdata, self.n_neighbors)
        if self._distances is not None:
            sg._distances = self._distances.__getitem__(*args)
        if self._neighbors is not None:
            sg._neighbors = self._neighbors.__getitem__(*args)
        if self._neighbor_types is not None:
            sg._neighbor_types = self._neighbor_types.__getitem__(*args)

    def update_knn(self, n_neighbors, re_run=True):
        """update_knn: Updates the KNN representation that gives rise to self.neighbors, self.distances and self.neighbor_types. This does not happen automatically after slicing, so it should be performed regularity before any graph-based analysis. Can be memory-intensive.

        :param n_neighbors: Number of neighbors for each neighbor graph.
        :type n_neighbors: int
        :param re_run: re-run, or use current representation, defaults to True
        :type re_run: bool, optional
        """

        if self._neighbors is not None and (n_neighbors <
                                            self._neighbors.shape[1]):
            self.n_neighbors = n_neighbors
            # return (self._neighbors[:, :n_neighbors],
            #         self._distances[:, :n_neighbors],
            #         self._neighbor_types[:, :n_neighbors])
        else:

            coordinates = np.stack([self.sdata.x, self.sdata.y]).T
            knn = NearestNeighbors(n_neighbors=n_neighbors)
            knn.fit(coordinates)
            self._distances, self._neighbors = knn.kneighbors(coordinates)
            self._neighbor_types = np.array(self.sdata.gene_ids)[
                self._neighbors]

            self.n_neighbors = n_neighbors

            # return self.distances, self.neighbors, self.neighbor_types

    def knn_entropy(self, n_neighbors=4):

        self.update_knn(n_neighbors=n_neighbors)
        indices = self.neighbors  # (n_neighbors=n_neighbors)

        knn_cells = np.zeros_like(indices)
        for i in range(indices.shape[1]):
            knn_cells[:, i] = self.sdata['gene_id'].iloc[indices[:, i]]

        H = np.zeros((len(self.sdata.genes), ))

        for i, gene in enumerate(self.sdata.genes):
            x = knn_cells[self.sdata['gene_id'] == i]
            _, n_x = np.unique(x[:, 1:], return_counts=True)
            p_x = n_x / n_x.sum()
            h_x = -(p_x * np.log2(p_x)).sum()
            H[i] = h_x

        return (H)

    def plot_entropy(self, n_neighbors=4):

        H = self.knn_entropy(n_neighbors)

        idcs = np.argsort(H)
        plt.figure(figsize=(25, 25))

        fig, axd = plt.subplot_mosaic([
            ['scatter_1', 'scatter_2', 'scatter_3', 'scatter_4'],
            ['bar', 'bar', 'bar', 'bar'],
            ['scatter_5', 'scatter_6', 'scatter_7', 'scatter_8'],
        ],
            figsize=(11, 7),
            constrained_layout=True)

        dem_plots = np.array([
            0,
            2,
            len(H) - 3,
            len(H) - 1,
            1,
            int(len(H) / 2),
            int(len(H) / 2) + 1,
            len(H) - 2,
        ])
        colors = ('royalblue', 'goldenrod', 'red', 'purple', 'lime',
                  'turquoise', 'green', 'yellow')

        axd['bar'].bar(
            range(len(H)),
            H[idcs],
            color=[
                colors[np.where(
                    dem_plots == i)[0][0]] if i in dem_plots else 'grey'
                for i in range(len(self.sdata.stats.counts))
            ])

        axd['bar'].set_xticks(range(len(H)),
                              [self.sdata.genes[h] for h in idcs],
                              rotation=90)
        axd['bar'].set_ylabel('knn entropy, k=' + str(n_neighbors))

        for i in range(8):
            idx = idcs[dem_plots[i]]
            gene = self.sdata.genes[idx]
            plot_name = 'scatter_' + str(i + 1)
            axd[plot_name].set_title(gene)
            axd[plot_name].scatter(
                self.sdata.x, self.sdata.y, color=(0.5, 0.5, 0.5, 0.1))
            axd[plot_name].scatter(self.sdata.x[self.sdata['gene_id'] == idx],
                                   self.sdata.y[self.sdata['gene_id'] == idx],
                                   color=colors[i],
                                   marker='.')
            axd[plot_name].set_xticks([], [])
            # if i>0:
            axd[plot_name].set_yticks([], [])

            if i < 4:
                y_ = (H[idcs])[i]
                _y = 0
            else:
                y_ = 0
                _y = 1

            print(axd)

            con = ConnectionPatch(xyA=(dem_plots[i], y_),
                                  coordsA=axd['bar'].transData,
                                  xyB=(np.mean(axd[plot_name].get_xlim()),
                                       axd[plot_name].get_ylim()[_y]),
                                  coordsB=axd[plot_name].transData,
                                  color='white',
                                  linewidth=1,
                                  linestyle='dotted')
            fig.add_artist(con)

    def _determine_counts(self, bandwidth=1, regularization=0, kernel=None):
        """_determine_counts: Determines the count distributions of genes around each molecule, thus effectively generating models of the immediate environment. 

        :param bandwidth: Bandwidth of the default Gaussian kernel, defaults to 1
        :type bandwidth: int, optional
        :param kernel: callable kernel function. If none is provided, a Gaussian function is used. Defaults to None
        :type kernel: callable, optional
        :return: 2d array of count distributions
        :rtype: np.Array
        """

        counts = np.zeros((len(self.sdata,), len(self.sdata.genes)))
        if kernel is None:
            def kernel(x): return np.exp(-x**2/(2*bandwidth**2))

        for i in range(0, self.n_neighbors):
            counts[np.arange(len(self.sdata)), self.neighbor_types[:, i]
                   ] += kernel(self.distances[:, i])

        assert (all(counts.sum(1)) > 0)
        counts[np.arange(len(self.sdata)),
               self.sdata.gene_ids] += regularization-1

        return counts

    def run_umap(self, bandwidth=1, kernel=None, metric='euclidean', zero_weight=1.0, cutoff=None, random_state=None, *args, **kwargs):
        """run_umap: Creates a UMAP representation of recurring local contexts in the source data.

        :param bandwidth: Bandwidth of the default Gaussian kernel used to build local environment models, defaults to 1
        :type bandwidth: int, optional
        :param kernel: Callable function to generate a model of the local spatial context for each spot's KNN graph. If none is provided, the default Gaussian is used. Defaults to None
        :type kernel: callable, optional
        :param metric:  Kernel description used as a UMAP parameter, defaults to 'euclidean'
        :type metric: str, callable, optional
        :param zero_weight: Regularization parameter that adds information of each spot's gene label to its local context model. High value encourages spots to form clusters with spots of their own gene label, defaults to 1
        :type zero_weight: float, optional
        :param random_state: Random state to initialize UMAP and FastICA.
        :type random_state: int
        """
        # print(kwargs)
        counts = self._determine_counts(
            bandwidth=bandwidth, kernel=kernel, regularization=zero_weight)

        if cutoff is not None:

            # print(f'PCA redution to {cutoff} dimensions.')
            # pca=PCA()
            # facs = pca.fit_transform(counts)

            # counts = np.zeros((counts.shape[0],cutoff+10))

            # counts[:,:cutoff] = facs[:,:cutoff]

            print(f'Reducing dimensions with FastICA')
            ica = FastICA(n_components=cutoff,random_state=random_state)
            counts = ica.fit_transform(counts)

            # counts[:,cutoff:] = facs
            # del facs,ica
            print('Calculating UMAP embedding.')
            # counts[:,-1]=facs[:,cutoff:].sum(1)

        umap = UMAP(metric=metric, random_state=random_state, *args, **kwargs)
        self._umap = umap.fit_transform(counts)

    def plot_umap(self, color_prop='genes', text_prop=None,
                  text_color_prop=None, c=None, color=None, color_dict=None, text_distance=1,
                  thlds_text=(1.0, 0.0, 0.0), text_kwargs={}, legend=False, **kwargs):

        embedding = self.umap

        categories = self.sdata.var[color_prop].unique()
        colors = self.sdata.var.project('c_'+color_prop)

        if color is not None:
            colors = (color,)*len(self.sdata)
        if c is not None:
            colors = c

        if legend:
            handlers = [plt.scatter([], [], color=self.sdata.var[self.sdata.var[color_prop]
                                    == c]['c_'+color_prop][0]) for c in sorted(categories)]
            plt.legend(handlers, sorted(categories))

        plt.scatter(*embedding.T, c=colors, **kwargs)

        if text_prop is not None:

            text_xs = []
            text_ys = []
            text_cs = []
            text_is = []
            text_zs = []

            X, Y = np.mgrid[embedding[:, 0].min():embedding[:, 0].max():100j,
                            embedding[:, 1].min():embedding[:, 1].max():100j]
            positions = np.vstack([X.ravel(), Y.ravel()])

            for i, g in enumerate(self.sdata.var[text_prop].unique()):
                if text_color_prop is not None:
                    fontcolor = self.sdata.var[self.sdata.var[text_prop]
                                                == g]['c_'+text_color_prop][0]
                else:
                    fontcolor = 'w'

                # try:
                embedding_subset = embedding[self.sdata.g.isin(
                    self.sdata.var[self.sdata.var[text_prop] == g].index)].T
                if embedding_subset.shape[1] > 2:
                    kernel = scipy.stats.gaussian_kde(embedding_subset)

                    Z = np.reshape(kernel(positions).T, X.shape)
                    localmaxs = scipy.ndimage.maximum_filter(Z, size=(10, 10))
                    maxz = (localmaxs == Z) & (localmaxs >=
                                               (localmaxs.max()*thlds_text[0]))
                    maxs = np.where(maxz)
                    # maxz = Z.max()
                    maxx = X[:, 0][maxs[0]]
                    maxy = Y[0][maxs[1]]

                    for j in range(len(maxx)):
                        # plt.suptitle(maxx[j])
                        text_xs.append(maxx[j])
                        text_ys.append(maxy[j])
                        text_is.append(g)
                        text_cs.append(fontcolor)
                        text_zs.append(Z[maxs[0][j], maxs[1][j]])

            cogs = self._untangle_text(
                np.array([text_xs, text_ys, ]).T, min_distance=text_distance)
            for i, c in enumerate(cogs):
                # plt.suptitle(text_zs)
                if (text_zs[i] > (max(text_zs)*thlds_text[1])):
                    if (thlds_text[2] is not None) or ((text_is[i] in self.sdata.counts) and (self.sdata.counts[text_is[i]] > thlds_text[2])):
                        txt = plt.text(
                            c[0], c[1], text_is[i], color=text_cs[i], ha='center', **text_kwargs)
                        txt.set_path_effects(
                            [PathEffects.withStroke(linewidth=3, foreground='k')])

                # except:
                #     print(f'Failed to assign text label to {g}')
    def umap_js(self, color_prop='c_genes'):
        """umap_js: A javascript-based display and selection function for the spatial source data and generated UMAP embedding. Can be used to understand UMAP agglomerations by displaying their gene compositions and projecting them back onto the original source data coordinates.

        :param color_prop: Property by which to color the spots in the scatter plots of the source data coordinates as well as the UMAP representation. Needs to be a column in sdata.var, defaults to 'c_genes'
        :type color_prop: str, optional
        :return: plotly HTML/javascript widget that displays in the jupyter notebook.
        :rtype: plotly.Widget
        """

        n_bars = 20

        trace_scatter = go.Scattergl(x=self.sdata.x,
                                        y=self.sdata.y,
                                        mode='markers',
                                        marker=dict(
                                            color=self.sdata.var.project(color_prop)),
                                        hoverinfo='none', meta={'name': 'tissue-scatter'},
                                        unselected={'marker': {
                                            'color': 'black', 'opacity': 0.2}},
                                        )

        f_scatter = go.FigureWidget(trace_scatter,)

        f_umap = go.FigureWidget(go.Scattergl(x=self.sdata.graph.umap[:, 0],
                                              y=self.sdata.graph.umap[:, 1],
                                              mode='markers',
                                              marker=dict(
                                                  color=self.sdata.var.project(color_prop)),
                                              unselected={'marker': {
                                                  'color': 'black', 'opacity': 0.2}},
                                              hoverinfo='none', meta={'name': 'umap-scatter'}),
                                 )

        colors = self.sdata.var.loc[self.sdata.stats.sort_values(
            'counts')[-n_bars:].index, color_prop].values

        w_bars = go.Bar(x=self.sdata.stats.sort_values('counts')[-n_bars:].index,
                        y=self.sdata.stats.sort_values(
                            'counts')[-n_bars:]['counts'],
                        marker={
                            'color': ['rgb'+str(tuple((np.array(c)*256).astype(int))) for c in colors]},
                        )
        f_bars = go.FigureWidget(w_bars,
                                 layout={'title': 'n (selected):'},
                                 )

        f_bars.data[0]['showlegend'] = False

        w_bars_ratio_up = go.Bar(x=self.sdata.stats.sort_values('counts')[-n_bars:].index,
                                 y=[0]*n_bars,
                                 marker={
                                     'color': ['rgb'+str(tuple((np.array(c)*256).astype(int))) for c in colors]},
                                 )
        f_bars_ratio_up = go.FigureWidget(w_bars_ratio_up,
                                          layout={'title': 'n (selected) / total:'})

        f_bars_ratio_up.data[0]['showlegend'] = False

        w_bars_binom = go.Bar(x=self.sdata.stats.sort_values('counts')[-n_bars:].index,
                              y=[0]*n_bars,
                              marker={
                                  'color': ['rgb'+str(tuple((np.array(c)*256).astype(int))) for c in colors]},
                              )
        f_bars_binom = go.FigureWidget(w_bars_binom,
                                       layout={
                                           'title': 'binomcdf (n (selected),p(total)):'},)

        f_bars_binom.data[0]['showlegend'] = False

        out = widgets.Output(layout={'border': '1px solid black'})

        def update_bars(plot, points, selector):

            if plot['meta']['name'] == 'tissue-scatter':
                f_umap.data[0].selectedpoints = plot.selectedpoints

            else:
                print(f_scatter.data)
                f_scatter.data[-1].selectedpoints = plot.selectedpoints

            subset = self.sdata[np.array(points.point_inds)]

            colors = subset.var.loc[subset.stats.sort_values(
                'counts')[-n_bars:].index, color_prop].values
            colors = ['rgb'+str(tuple((np.array(c)*256).astype(int)))
                      for c in colors]
            ys = subset.stats.sort_values('counts')[-n_bars:]['counts']
            xs = subset.stats.sort_values('counts')[-n_bars:].index

            f_bars.data[0].marker.color = colors
            f_bars.data[0].x = xs
            f_bars.data[0].y = ys

            vals = (subset.stats.counts) / \
                (self.sdata.stats.loc[subset.stats.index].counts)
            idcs = np.argsort(vals)

            colors = subset.var[color_prop][idcs]
            colors = ['rgb'+str(tuple((np.array(c)*256).astype(int)))
                      for c in colors]
            ys = vals[idcs]
            xs = subset.stats.index[idcs]

            f_bars_ratio_up.data[0].marker.color = colors[-n_bars:]
            f_bars_ratio_up.data[0].x = xs[-n_bars:]
            f_bars_ratio_up.data[0].y = (ys[-n_bars:])

            vals = binom.cdf(subset.stats.counts, self.sdata.stats.loc[subset.stats.index].counts, len(
                subset)/len(self.sdata))
            idcs = np.argsort(vals)

            colors = subset.var[color_prop][idcs]
            colors = ['rgb'+str(tuple((np.array(c)*256).astype(int)))
                      for c in colors]
            ys = vals[idcs]
            xs = subset.stats.index[idcs]

            f_bars_binom.data[0].marker.color = colors[-n_bars:]
            f_bars_binom.data[0].x = xs[-n_bars:]
            f_bars_binom.data[0].y = (ys[-n_bars:])

        f_scatter.data[-1].on_selection(update_bars)
        f_umap.data[0].on_selection(update_bars)

        text_field = widgets.Text(
            value='selection1',
            placeholder='Label for storing',
            description='Name:',
            disabled=False
        )

        store_button = widgets.Button(
            description='store selection',
            disabled=False,
            button_style='',  # 'success', 'info', 'warning', 'danger' or ''
        )

        def store_selection(event):
            self.sdata[text_field.value] = (pd.Series(
                np.arange(len(self.sdata))).isin(f_umap.data[0].selectedpoints)).values

        store_button.on_click(store_selection)

        reset_button = widgets.Button(
            description='reset selection',
            disabled=False,
            button_style='',  # 'success', 'info', 'warning', 'danger' or ''
        )

        return widgets.VBox([widgets.HBox([f_scatter, f_umap], layout=Layout(display='flex', width='100%', height='80%', border='red solid 1px', align_items='stretch', justify_content='space-around', flex_direction='row')),
                            widgets.HBox([widgets.HBox([f_bars, f_bars_ratio_up, f_bars_binom], layout=Layout(display='flex', width='80%')),
                                          widgets.VBox([widgets.HBox([text_field,
                                                                      ]), widgets.HBox([store_button, reset_button])], layout=Layout(display='flex', width='20%', height='100%', border='red solid 1px', justify_content='space-around', flex_direction='column')
                                                       )]
                                         )], layout=Layout(width='100%', height='80vh', background='red', border='solid 1px'))

    def map_and_umap(self, color_prop=None, scalebar=True, cmap='jet',
                     **kwargs):
        """map_and_umap: Plots a side-by-side representation of the available UMAP- and coordinate data, with styling arguments passed to both plotting functions. 

        :param color_prop: Property to color the individual markers by. Needs to be a column in self.sdata.var, defaults to None
        :type color_prop: _type_, optional
        :param scalebar: Whether to display a scalebar in the coordinate representation, defaults to True
        :type scalebar: bool, optional
        """

        if color_prop is None:
            color_prop = 'genes'
        else:
            kwargs['color_prop'] = color_prop
            # scatter_kwargs['color_prop'] = color_prop

        plt.style.use('dark_background')

        fig = plt.gcf()

        ax1 = plt.subplot2grid((3, 2), (0, 0), 2, 1)

        sc2, _, _ = self.sdata.scatter(
            axd=ax1, scalebar=scalebar, cmap=cmap,  ** kwargs)

        ax2 = plt.subplot2grid((3, 2), (0, 1), 2, 1)
        self.sdata.graph.plot_umap(cmap=cmap, **kwargs)

    def _untangle_text(self, cogs, untangle_rounds=50, min_distance=0.5):
        knn = NearestNeighbors(n_neighbors=2)

        np.random.seed(42)
        cogs_new = cogs+np.random.normal(size=cogs.shape,)*0.01

        for i in range(untangle_rounds):

            cogs = cogs_new.copy()
            knn = NearestNeighbors(n_neighbors=2)

            knn.fit(cogs)
            distances, neighbors = knn.kneighbors(cogs/[1.01, 1.0])
            too_close = (distances[:, 1] < min_distance)

            for i, c in enumerate(np.where(too_close)[0]):
                partner = neighbors[c, 1]
                cog = cogs[c]-cogs[partner]
                cog_new = cogs[c]+0.3*cog
                cogs_new[c] = cog_new

        return cogs_new



from importlib.resources import path
from textwrap import fill
from multiprocessing import Pool

import numpy as np
import community
import networkx as nx

# from requests import patch
from scipy.ndimage import gaussian_filter, maximum_filter
from plankton.pixelmaps import PixelMap

from sklearn.neighbors import NearestNeighbors,kneighbors_graph
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans,DBSCAN
from sklearn import preprocessing

from skimage import filters

import matplotlib.pyplot as plt

from tqdm import tqdm_notebook
import scipy
from scipy.spatial.distance import cdist

def determine_gains(stats1, stats2):

    norm_counts1 = stats1.counts/stats1.counts.sum()
    norm_counts2 = stats2.counts/stats2.counts.sum()

    return norm_counts1/norm_counts2


def hbar_compare(stat1, stat2, labels=None, text_display_threshold=0.02, c=None):

    genes_united = sorted(
        list(set(np.concatenate([stat1.index, stat2.index]))))[::-1]
    counts_1 = [
        0]+[stat1.loc[i].counts if i in stat1.index else 0 for i in genes_united]
    counts_2 = [
        0]+[stat2.loc[i].counts if i in stat2.index else 0 for i in genes_united]
    cum1 = np.cumsum(counts_1)/sum(counts_1)
    cum2 = np.cumsum(counts_2)/sum(counts_2)

    if c is None:
        c = [None]*len(cum1)

    for i in range(1, len(cum1)):

        bars = plt.bar([0, 1], [cum1[i]-cum1[i-1], cum2[i]-cum2[i-1]],
                       bottom=[cum1[i-1], cum2[i-1], ], width=0.4, color=c[i-1])
        clr = bars.get_children()[0].get_facecolor()
        plt.plot((0.2, 0.8), (cum1[i], cum2[i]),
                 c=plt.rcParams['axes.facecolor'], alpha=0.7)
        plt.fill_between(
            (0.2, 0.8), (cum1[i], cum2[i]), (cum1[i-1], cum2[i-1]), color=clr, alpha=0.2)

        if (counts_1[i]/sum(counts_1) > text_display_threshold) or \
                (counts_2[i]/sum(counts_2) > text_display_threshold):
            plt.text(0.5, (cum1[i]+cum1[i-1]+cum2[i]+cum2[i-1])/4,
                     genes_united[i-1], ha='center',)

    if labels is not None:
        plt.xticks((0, 1), labels)


def sorted_bar_compare(stat1, stat2, kwargs1={}, kwargs2={}):
    categories_1 = (stat1.index)
    counts_1 = (np.array(stat1.counts).flatten())
    counts_1_idcs = np.argsort(counts_1)
    # count_ratios = np.log(self.determine_gains())
    # count_ratios -= count_ratios.min()
    # count_ratios /= count_ratios.max()

    ax1 = plt.subplot(311)
    ax1.set_title('compared     lecule counts:')
    ax1.bar(np.arange(len(counts_1)),
            counts_1[counts_1_idcs], color='grey', **kwargs1)
    # ax1.set_ylabel('log(count) scRNAseq')
    ax1.set_xticks(np.arange(len(categories_1)),
                   categories_1[counts_1_idcs],
                   )
    ax1.tick_params(axis='x', rotation=90)
    ax1.set_yscale('log')

    ax2 = plt.subplot(312)
    for i, gene in enumerate(categories_1[counts_1_idcs]):
        if gene in stat2.index:
            plt.plot(
                [i, stat2.count_ranks[gene]],
                [1, 0],
            )
    plt.axis('off')
    ax2.set_ylabel(' ')

    ax3 = plt.subplot(313)
    ax3.bar(np.arange(len(stat2)),
            stat2.counts[stat2.count_indices], color='grey', **kwargs2)

    ax3.set_xticks(np.arange(len(stat2.index)),
                   stat2.index[stat2.count_indices],
                   rotation=90)
    ax3.set_yscale('log')
    ax3.invert_yaxis()
    ax3.xaxis.tick_top()
    ax3.xaxis.set_label_position('top')
    # ax3.set_ylabel('log(count) spatial')
    return(ax1, ax2, ax3)


def fill_celltypemaps(ct_map, fill_blobs=True, min_blob_area=0, filter_params={}, output_mask=None):
    """
    Post-filter cell type maps created by `map_celltypes`.

    :param min_r: minimum threshold of the correlation.
    :type min_r: float
    :param min_norm: minimum threshold of the vector norm.
        If a string is given instead, then the threshold is automatically determined using
        sklearn's `threshold filter functions <https://scikit-image.org/docs/dev/api/skimage.filters.html>`_ (The functions start with `threshold_`).
    :type min_norm: str or float
    :param fill_blobs: If true, then the algorithm automatically fill holes in each blob.
    :type fill_blobs: bool
    :param min_blob_area: The blobs with its area less than this value will be removed.
    :type min_blob_area: int
    :param filter_params: Filter parameters used for the sklearn's threshold filter functions.
        Not used when `min_norm` is float.
    :type filter_params: dict
    :param output_mask: If given, the cell type maps will be filtered using the output mask.
    :type output_mask: np.ndarray(bool)
    """

    from skimage import measure

    filtered_ctmaps = np.zeros_like(ct_map) - 1

    for cidx in np.unique(ct_map):
        mask = ct_map == cidx
        if min_blob_area > 0 or fill_blobs:
            blob_labels = measure.label(mask, background=0)
            for bp in measure.regionprops(blob_labels):
                if min_blob_area > 0 and bp.filled_area < min_blob_area:
                    for c in bp.coords:
                        mask[c[0], c[1], ] = 0

                    continue
                if fill_blobs and bp.area != bp.filled_area:
                    minx, miny,  maxx, maxy, = bp.bbox
                    mask[minx:maxx, miny:maxy, ] |= bp.filled_image

        filtered_ctmaps[np.logical_and(mask == 1, np.logical_or(
            ct_map == -1, ct_map == cidx))] = cidx

    return filtered_ctmaps


def get_histograms(sdata, mins=None, maxs=None, category=None, resolution=5):

    if mins is None:
        mins = sdata.coordinates.min(0)

    if maxs is None:
        maxs = sdata.coordinates.max(0)+1

    # print(mins,maxs)

    n_bins = np.ceil(np.divide(maxs-mins, resolution)).astype(int)

    histograms = []

    if category is None:
        for gene in sdata.genes:
            histograms.append(np.histogram2d(
                *sdata[sdata.g == gene].coordinates.T, bins=n_bins, range=([mins[0], maxs[0]], [mins[1], maxs[1]]))[0])

    else:
        for c in sdata[category].cat.categories:
            histograms.append(np.histogram2d(
                *sdata[sdata[category] == c].coordinates.T, bins=n_bins, range=([mins[0], maxs[0]], [mins[1], maxs[1]]))[0])

    return histograms,


def crosscorr(x, y):
    x -= np.array(x.mean(1))[:, None]
    y -= np.array(y.mean(1))[:, None]
    c = (np.dot(x, y.T)/x.shape[1]).squeeze()

    return np.nan_to_num(np.nan_to_num(c/np.array(x.std(1))[:, None])/np.array(y.std(1))[None, :])


def ssam(sdata, signatures=None, adata_obs_label='celltype', kernel_bandwidth=2.5, output_um_p_px=5,
         patch_length=1000, threshold_exp=0.1, threshold_cor=0.1, background_value=-1,
         fill_blobs=True, min_blob_area=10):

    if signatures is None:
        signatures = sdata.uns['ssam']['centroids']

    # if (sdata.scanpy is not None):
        # signatures = sdata.scanpy.generate_signatures(adata_obs_label)

    kernel_bandwidth_px = kernel_bandwidth/output_um_p_px

    out_shape = (np.ceil(sdata.x.max()/output_um_p_px+kernel_bandwidth_px*3).astype(int),
                 np.ceil(sdata.y.max()/output_um_p_px+kernel_bandwidth_px*3).astype(int))

    ct_map = np.zeros((out_shape), dtype=int)+background_value
    vf_norm = np.zeros_like(ct_map)

    range_x = np.floor(sdata.x.min()).astype(
        int), np.ceil(sdata.x.max()).astype(int)
    patch_delimiters_x = list(
        range(range_x[0], range_x[1], patch_length))+[range_x[1]]

    range_y = np.floor(sdata.y.min()).astype(
        int), np.ceil(sdata.y.max()).astype(int)
    patch_delimiters_y = list(
        range(range_y[0], range_y[1], patch_length))+[range_y[1]]

    print(list(patch_delimiters_x), list(patch_delimiters_y))

    with tqdm_notebook(total=(len(patch_delimiters_x)-1)*(len(patch_delimiters_y)-1)) as pbar:
        for i, x in enumerate(patch_delimiters_x[:-1]):
            for j, y in enumerate(patch_delimiters_y[:-1]):
                sdata_patch = sdata.raw().spatial[x:patch_delimiters_x[i+1],
                                                y:patch_delimiters_y[j+1]]

                if not len(sdata_patch):
                    break

                mins = sdata_patch.coordinates.min(0).astype(int)
                maxs = sdata_patch.coordinates.max(
                    0).astype(int)+kernel_bandwidth*3
                hists = get_histograms(
                    sdata_patch, mins=mins, maxs=maxs, resolution=output_um_p_px)
                hists = np.concatenate(
                    [gaussian_filter(h, kernel_bandwidth_px) for h in hists])

                # print(hists.shape)

                norm = np.sum(hists, axis=0)

                vf_norm[(x+mins[0])//output_um_p_px:(x+mins[0])//output_um_p_px+norm.shape[0],
                        (y+mins[1])//output_um_p_px:(y+mins[1])//output_um_p_px+norm.shape[1], ] = norm

                mask = norm > threshold_exp

                exps = np.zeros((len(sdata.genes), mask.sum()))

                # print(exps.shape,signatures.shape)
                exps[sdata.stats.loc[sdata_patch.genes].gene_ids.values,
                    :] = hists[:, mask]

                local_ct_map = mask.astype(int)-1

                corrs = crosscorr(exps.T, signatures)
                # print(corrs.min())
                corrs_winners = corrs.argmax(1)
                corrs_winners[corrs.max(1) < threshold_cor] = background_value
                local_ct_map[mask] = corrs_winners

                # idcs=sdata_patch.index

                x_ = int(max(0, (x+mins[0])-kernel_bandwidth_px*3))//output_um_p_px
                y_ = int(max(0, (y+mins[1])-kernel_bandwidth_px*3))//output_um_p_px

                ct_map[x_:x_+local_ct_map.shape[0],
                    y_:y_+local_ct_map.shape[1]] = local_ct_map

                pbar.update(1)


    if not 'ssam' in sdata.uns.keys():
        sdata.uns['ssam']={}

    sdata.uns['ssam']['ct_map_raw']=ct_map
    sdata.uns['ssam']['vf_norm']=vf_norm

    ct_map = fill_celltypemaps(
        ct_map, min_blob_area=min_blob_area, fill_blobs=fill_blobs)


    sdata.uns['ssam']['ct_map_filtered']=ct_map

    return PixelMap(ct_map.T, px_p_um=1/output_um_p_px)


def localmax_sampling(sdata, min_distance=3, bandwidth=4):
    n_bins = np.array(sdata.coordinates.max(0)).astype(int)+2

    print(n_bins)
    vf = (gaussian_filter(np.histogram2d(
        *sdata.coordinates.T, bins=(range(n_bins[0]),range(n_bins[1])))[0], bandwidth))

    # return vf
    localmaxs = np.where((maximum_filter(vf, min_distance) == vf) & (vf > 0.2))
    knn = NearestNeighbors(n_neighbors=150)
    knn.fit(sdata.coordinates)
    dists, nbrs = knn.kneighbors(np.array(localmaxs).T)
    neighbor_types = np.array(sdata.gene_ids)[nbrs]

    localmax_samples = np.zeros((dists.shape[0], len(sdata.genes)))

    bandwidth = bandwidth
    def kernel(x): return np.exp(-x**2/(2*bandwidth**2))

    for i in range(0, dists.shape[1]):
        localmax_samples[np.arange(dists.shape[0]), neighbor_types[:, i]
               ] += kernel(dists[:, i])

    if not 'ssam' in sdata.uns.keys():
        sdata.uns['ssam']={}

    sdata.uns['ssam']['localmax_coords']=localmaxs
    sdata.uns['ssam']['localmax_samples']=localmax_samples

    return localmax_samples,localmaxs

def normalize_vectors(sdata, use_expanded_vectors=False, normalize_gene=False, normalize_vector=False, normalize_median=False, 
                        norm='l1',size_after_normalization=1e4, log_transform=False, scale=False):
    """
    Normalize and regularize vectors

    :param use_expanded_vectors: If True, use averaged vectors nearby local maxima of the vector field.
    :type use_expanded_vectors: bool
    :param normalize_gene: If True, normalize vectors by sum of each gene expression across all vectors.
    :type normalize_gene: bool
    :param normalize_vector: If True, normalize vectors by sum of all gene expression of each vector.
    :type normalize_vector: bool
    :param log_transform: If True, vectors are log transformed.
    :type log_transform: bool
    :param scale: If True, vectors are z-scaled (mean centered and scaled by stdev).
    :type scale: bool
    """
    if use_expanded_vectors:
        vec = np.array(sdata.uns['ssam']['expanded_vectors'], copy=True)
    else:
        vec = np.array(sdata.uns['ssam']['localmax_samples'], copy=True)
    if normalize_gene:
        vec = preprocessing.normalize(vec, norm=norm, axis=0) * size_after_normalization  # Normalize per gene
    if normalize_vector:
        vec = preprocessing.normalize(vec, norm="l1", axis=1) * size_after_normalization # Normalize per vector
    if normalize_median:
        def n(v):
            s, m = np.sum(v, axis=1), np.median(v, axis=1)
            s[m > 0] = s[m > 0] / m[m > 0]
            s[m == 0] = 0
            v[s > 0] = v[s > 0] / s[s > 0][:, np.newaxis]
            v[v == 0] = 0
            return v
        vec = n(vec)
    if log_transform:
        vec = np.log2(vec + 1)
    if scale:
        vec = preprocessing.scale(vec)
    sdata.uns['ssam']['normalized_vectors'] = vec
    return

def cluster_vectors(sdata, pca_dims=10, min_cluster_size=0, resolution=0.6, prune=1.0/15.0, snn_neighbors=30, max_correlation=1.0,
                    metric="correlation", subclustering=False, dbscan_eps=0.4, centroid_correction_threshold=0.8, random_state=0):
    """
    Cluster the given vectors using the specified clustering method.

    :param pca_dims: Number of principal componants used for clustering.
    :type pca_dims: int
    :param min_cluster_size: Set minimum cluster size.
    :type min_cluster_size: int
    :param resolution: Resolution for Louvain community detection.
    :type resolution: float
    :param prune: Threshold for Jaccard index (weight of SNN network). If it is smaller than prune, it is set to zero.
    :type prune: float
    :param snn_neighbors: Number of neighbors for SNN network.
    :type snn_neighbors: int
    :param max_correlation: Clusters with higher correlation to this value will be merged.
    :type max_correlation: bool
    :param metric: Metric for calculation of distance between vectors in gene expression space.
    :type metric: str
    :param subclustering: If True, each cluster will be clustered once again with DBSCAN algorithm to find more subclusters.
    :type subclustering: bool
    :param centroid_correction_threshold: Centroid will be recalculated with the vectors
        which have the correlation to the cluster medoid equal or higher than this value.
    :type centroid_correction_threshold: float
    :param random_state: Random seed or scikit-learn's random state object to replicate the same result
    :type random_state: int or random state object
    """
    
    vecs_normalized = sdata.uns['ssam']['normalized_vectors']
    vecs_normalized_dimreduced = PCA(n_components=pca_dims, random_state=random_state).fit_transform(vecs_normalized)

    print(vecs_normalized_dimreduced.shape)

    def cluster_vecs(vecs):
        k = min(snn_neighbors, vecs.shape[0])
        knn_graph = kneighbors_graph(vecs, k, mode='connectivity', include_self=True, metric=metric).todense()
        intersections = np.dot(knn_graph, knn_graph.T)
        snn_graph = intersections / (k + (k - intersections)) # borrowed from Seurat
        snn_graph[snn_graph < prune] = 0
        G = nx.from_numpy_matrix(snn_graph)
        partition = community.best_partition(G, resolution=resolution, random_state=random_state)
        lbls = np.array(list(partition.values()))
        return lbls

    def remove_small_clusters(lbls, lbls2=None):
        small_clusters = []
        cluster_indices = []
        lbls = np.array(lbls)
        for lbl in np.unique(lbls):
            if lbl == -1:
                continue
            cnt = np.sum(lbls == lbl)
            if cnt < min_cluster_size:
                small_clusters.append(lbl)
            else:
                cluster_indices.append(lbl)
        for lbl in small_clusters:
            lbls[lbls == lbl] = -1
        tmp = np.array(lbls, copy=True)
        for i, idx in enumerate(cluster_indices):
            lbls[tmp == idx] = i
        if lbls2 is not None:
            for lbl in small_clusters:
                lbls2[lbls2 == lbl] = -1
            tmp = np.array(lbls2, copy=True)
            for i, idx in enumerate(cluster_indices):
                lbls2[tmp == idx] = i
            return lbls, lbls2
        else:
            return lbls
    
    if subclustering:
        super_lbls = cluster_vecs(vecs_normalized_dimreduced)
        dbscan = DBSCAN(eps=dbscan_eps, min_samples=min_cluster_size, metric=metric)
        all_lbls = np.zeros_like(super_lbls)
        global_lbl_idx = 0
        for super_lbl in set(list(super_lbls)):
            super_lbl_idx = np.where(super_lbls == super_lbl)[0]
            if super_lbl == -1:
                all_lbls[super_lbl_idx] = -1
                continue
            sub_lbls = dbscan.fit(vecs_normalized_dimreduced[super_lbl_idx]).labels_
            for sub_lbl in set(list(sub_lbls)):
                if sub_lbl == -1:
                    all_lbls[tuple([super_lbl_idx[sub_lbls == sub_lbl]])] = -1
                    continue
                all_lbls[tuple([super_lbl_idx[sub_lbls == sub_lbl]])] = global_lbl_idx
                global_lbl_idx += 1
    else:
        all_lbls = cluster_vecs(vecs_normalized_dimreduced)            
            
    new_labels = __correct_cluster_labels(vecs_normalized,all_lbls, centroid_correction_threshold)
    new_labels, all_lbls = remove_small_clusters(new_labels, all_lbls)
    centroids, centroids_stdev = __calc_centroid(vecs_normalized, new_labels)

    merge_candidates = []
    if max_correlation < 1.0:
        Z = scipy.cluster.hierarchy.linkage(centroids, metric='correlation')
        clbls = scipy.cluster.hierarchy.fcluster(Z, 1 - max_correlation, 'distance')
        for i in set(clbls):
            leaf_indices = np.where(clbls == i)[0]
            if len(leaf_indices) > 1:
                merge_candidates.append(leaf_indices)
        removed_indices = []
        for cand in merge_candidates:
            for i in cand[1:]:
                all_lbls[all_lbls == i] = cand[0]
                removed_indices.append(i)
        for i in sorted(removed_indices, reverse=True):
            all_lbls[all_lbls > i] -= 1

        new_labels = __correct_cluster_labels(vecs_normalized,all_lbls, centroid_correction_threshold)
        new_labels, all_lbls = remove_small_clusters(new_labels, all_lbls)
        centroids, centroids_stdev = __calc_centroid(vecs_normalized, new_labels)
            
    sdata.uns['ssam']['cluster_labels'] = all_lbls
    sdata.uns['ssam']['filtered_cluster_labels'] = new_labels
    sdata.uns['ssam']['centroids'] = np.array(centroids)
    sdata.uns['ssam']['centroids_stdev'] = np.array(centroids_stdev)
    #self.dataset.medoids = np.array(medoids)
    
    print("Found %d clusters"%len(centroids))
    return

def __correct_cluster_labels(normalized_vectors, cluster_labels, centroid_correction_threshold):
    new_labels = np.array(cluster_labels, copy=True)
    if centroid_correction_threshold < 1.0:
        for cidx in np.unique(cluster_labels):
            if cidx == -1:
                continue
            prev_midx = -1
            while True:
                vecs = normalized_vectors[new_labels == cidx]
                vindices = np.where(new_labels == cidx)[0]
                midx = vindices[np.argmin(np.sum(cdist(vecs, vecs), axis=0))]
                if midx == prev_midx:
                    break
                prev_midx = midx
                m = normalized_vectors[midx]
                for vidx, v in zip(vindices, vecs):
                    if crosscorr(v[None], m[None]) < centroid_correction_threshold:
                        new_labels[vidx] = -1
    return new_labels

def __calc_centroid(normalized_vectors, cluster_labels):
    centroids = []
    centroids_stdev = []
    #medoids = []
    for lbl in sorted(list(set(cluster_labels))):
        if lbl == -1:
            continue
        cl_vecs = normalized_vectors[cluster_labels == lbl, :]
        #cl_dists = scipy.spatial.distance.cdist(cl_vecs, cl_vecs, metric)
        #medoid = cl_vecs[np.argmin(np.sum(cl_dists, axis=0))]
        centroid = np.mean(cl_vecs, axis=0)
        centroid_stdev = np.std(cl_vecs, axis=0)
        #medoids.append(medoid)
        centroids.append(centroid)
        centroids_stdev.append(centroid_stdev)
    return centroids, centroids_stdev#, medoids




from pyexpat import XML_PARAM_ENTITY_PARSING_ALWAYS
from typing import Union
import collections

import numpy as np
import matplotlib.pyplot as plt


class PixelMap():

    def __init__(self,
                 pixel_data: np.ndarray,
                 px_p_um: float = 1.0,
                 x_shift=0.0,
                 y_shift=0.0,
                 cmap='Greys') -> None:

        self.data = pixel_data

        self.n_channels = 1 if len(
            pixel_data.shape) == 2 else pixel_data.shape[-1]

        if not isinstance(px_p_um, collections.abc.Iterable) or len(px_p_um) == 1:
            self.scale = (px_p_um, px_p_um)
        else:
            self.scale = px_p_um

        self.cmap=cmap

        self.extent = (x_shift, x_shift+pixel_data.shape[0] / self.scale[0],
                       y_shift, y_shift + pixel_data.shape[1] / self.scale[1])

    @property
    def shape(self):
        return self.extent[1] - self.extent[0], self.extent[3] - self.extent[2]

    def imshow(self, cmap=None, axd=None, **kwargs) -> None:
        extent = np.array(self.extent)

        if (len(self.data.shape) > 2) and (self.data.shape[2] > 4):
            data = self.data.sum(-1)
        else:
            data = self.data

        if axd is None:
            axd = plt.gca()

        if cmap is None:
            cmap = self.cmap

        axd.imshow(data, extent=extent[[0, 3, 1, 2]], cmap=cmap,**kwargs)

    def __getitem__(self, indices: Union[slice, collections.abc.Iterable[slice]]):

        # print(self.extent)
        if not isinstance(indices, collections.abc.Iterable):
            index_x = indices
            index_y = slice(0, None, None)

        else:
            index_x = indices[0]

            if len(indices) > 1:
                index_y = indices[1]
            else:
                index_y = slice(0, None, None)

        if (index_x.start is None):
            start_x = 0  # self.extent[0]
        else:
            start_x = index_x.start
        if (index_x.stop is None):
            stop_x = self.extent[3]-1
        else:
            stop_x = index_x.stop

        if (index_y.start is None):
            start_y = 0  # self.extent[2]
        else:
            start_y = index_y.start
        if (index_y.stop is None):
            stop_y = self.extent[1]-1
        else:
            stop_y = index_y.stop

        # print(self.data.shape, int(stop_x * self.scale[1]), int(stop_y * self.scale[0]), self.scale)
        data = self.data[int(start_y * self.scale[1]):int(stop_y *
                                                          self.scale[1]),
                         int(start_x * self.scale[0]):int(stop_x *
                                                          self.scale[0]), ]

        return PixelMap(
            data,
            px_p_um=self.scale,
        )

    def get_value(self, x, y, padding_value=-1):
        x_= np.array(y).flatten()
        y= np.array(x).flatten()
        x=x_
        x = np.round((x-self.extent[0])/self.extent[1]*self.data.shape[0]).astype(int)
        y = np.round((y-self.extent[2])/self.extent[3]*self.data.shape[1]).astype(int)


        values = np.empty_like(x)
        valid = (x>=0)&(y>=0)&(x<self.data.shape[0])&(y<self.data.shape[1])
        print(x    )
        values[~valid] = padding_value
        values[valid] = self.data[x[valid],y[valid]]
        return values





class KDEProjection(PixelMap):
    def __init__(self, sd,
                 bandwidth: float = 3.0,
                 threshold_vf_norm: float = 1.0,
                 threshold_p_corr: float = 0.5,
                 px_p_um: float = 1) -> None:

        self.sdata = sd
        self.bandwidth = bandwidth
        self.threshold_vf_norm = threshold_vf_norm
        self.threshold_p_corr = threshold_p_corr

        self.scale = px_p_um

        super().__init__(self.run_kde(), px_p_um)

    def run_kde(self) -> None:

        kernel = self.generate_kernel(self.bandwidth*3, self.scale)

        x_int = np.array(self.sdata.y * self.scale).astype(int)
        y_int = np.array(self.sdata.x * self.scale).astype(int)
        genes = self.sdata.gene_ids

        vf = np.zeros(
            (x_int.max()+kernel.shape[0]+1, y_int.max()+kernel.shape[0]+1, len(self.sdata.genes)))

        for x, y, g in zip(x_int, y_int, genes):
            # print(x,y,vf.shape,kernel.shape)
            vf[x:x+kernel.shape[0], y:y+kernel.shape[1], g] += kernel

        return vf[kernel.shape[0]//2:-kernel.shape[0]//2, kernel.shape[1]//2:-kernel.shape[1]//2]

    def generate_kernel(self, bandwidth: float, scale: float = 1) -> np.ndarray:

        kernel_width_in_pixels = int(bandwidth * scale *
                                     6)  # kernel is 3 sigmas wide.

        span = np.linspace(-3, 3, kernel_width_in_pixels)
        X, Y = np.meshgrid(span, span)

        return 1 / (2 * np.pi)**0.5 * np.exp(-0.5 * ((X**2 + Y**2)**0.5)**2)


class DensityMap(PixelMap):
    def __init__(self, data, *args, **kwargs):

        # .super().qd

        pass


class PixelMask(PixelMap):
    def __init__(self,
                 pixel_data: np.ndarray,
                 px_p_um: float = 1.0,
                 x_shift=0.0,
                 y_shift=0.0) -> None:

        super(PixelMask, self).__init__(
            pixel_data=pixel_data,
            px_p_um=px_p_um,
            x_shift=x_shift,
            y_shift=y_shift
        )

# decorate PixelMap to overload python operators:

def overload_operator(name):

    def apply(self,*args): 
        # getattr(self.data   , name)(*args)
        pixel_map_ = self[:]
        pixel_map_.data = getattr(self.data, name)(*args)
        return pixel_map_

    return apply
    # return lambda self, *args: PixelMap(getattr(self.data, name)(*args), self.info)

for name in ["__add__", "__sub__", "__mul__", "__truediv__", "__floordiv__", "__invert__", "__neg__", "__pos__",
            "__lt__","__le__","__eq__","__ne__","__ge__","__gt__",]:
    setattr(PixelMap, name, overload_operator(name))




from weakref import ref
from scipy.signal import fftconvolve
import numpy as np
from scipy import signal, fft as sp_fft


def _get_kernels(max_radius, linear_steps):

    if linear_steps > max_radius:
        linear_steps = max_radius

    vals = np.ones((max_radius)+1)
    vals[0] = 0
    vals[linear_steps+1:] += np.arange(max_radius-linear_steps)
    vals = np.cumsum(vals)
    vals = vals[:(vals < max_radius).sum()+1]
    vals[-1] = max_radius

    span = np.arange(-max_radius, max_radius + 1)
    X, Y = np.meshgrid(span, span)
    dists = (X**2+Y**2)**0.5

    kernel_1 = np.zeros_like(X)
    kernel_1[max_radius, max_radius] = 1
    kernels = [kernel_1]

    for i in range(len(vals)-1):

        r1 = vals[i]
        r2 = vals[i+1]

        kernel_1 = (dists-r1)
        kernel_1 = -(kernel_1-1)*(kernel_1 < 1)
        kernel_1[kernel_1 > 1] = 1
        kernel_1 = 1-kernel_1

        kernel_2 = (dists-r2)
        kernel_2 = -(kernel_2-1)*(kernel_2 < 1)
        kernel_2[kernel_2 > 1] = 1
        kernel_2[kernel_2 == 1] = kernel_1[kernel_2 == 1]

        kernels.append(kernel_2)

    kernels = np.array(kernels)

    return (kernels,vals)

def get_histograms(sdata, category=None,resolution=5, mins=None, maxs=None):

    if mins is None: mins = sdata.coordinates.min(0)
    if maxs is None: maxs = sdata.coordinates.max(0)

    n_bins = np.ceil(np.divide(maxs-mins, resolution)).astype(int)

    histograms=[]
    
    if category is None:
        for gene in sdata.genes:
            histograms.append(np.histogram2d(*sdata[sdata.g==gene].coordinates.T, bins=n_bins, range=([mins[0], maxs[0]], [mins[1], maxs[1]]))[0])
        
    else:
        for c in sdata[category].cat.categories:
            histograms.append(np.histogram2d(*sdata[sdata[category]==c].coordinates.T, bins=n_bins, range=([mins[0], maxs[0]], [mins[1], maxs[1]]))[0])

    return histograms

# import pdb;

def _interpolate(x,y,step):

    from scipy import interpolate 
    """_interpolat generates linear interpolation from the x,y to create mutual distance curves in um units

    :param x: distances with available values
    :type x: np.array
    :param y: values at x
    :type y: np array
    :param step: step size of x,y
    :type step: float
    """

    # return lambda x_ :  interpolate.BSpline(*interpolate.splrep(x,y,s=0,k=5), extrapolate=False)

    return np.apply_along_axis(lambda x_ :  interpolate.BSpline(*interpolate.splrep(x,x_,s=0,k=5), extrapolate=False)(np.arange(0,x.max()*step)),2,y)
    # return lambda x_ : np.apply_along_axis( Akima1DInterpolator(x,x_)(np.arange(0,x.max()*step)),2,y)
        # return np.apply_along_axis(lambda x_: np.interp(np.arange(0,x.max()*step),x*step,x_,),2,y)

def co_occurrence(sdata, resolution=5.0, max_radius=400, linear_steps=5, category=None):
    """co_occurrence _summary_

    :param sdata: SpatialData - all spots contained in sdata will be analysed. 
    :type sdata: SpatialData
    :param resolution: Smallest resolution of the co-occurence model in um, defaults to 5.
    :type resolution: float, optional
    :param max_radius: Largest radius of the co-occurrence model in um, defaults to 400.
    :type max_radius: float, optional
    :param linear_steps: Number of linear steps to model local heterogeniety. Afterwards, distance bins get wider to save computational resources, defaults to 5
    :type linear_steps: int, optional
    :param category: Category to model the co-occurrence of. Must be a column in 'sdata' with dtype 'category'. When given 'None', the algorithm defaults to gene labels.
    :type category: str, optional
    :return: _description_
    :rtype: _type_
    """

    hists = get_histograms(sdata, category=category)  

    if max_radius is None:
        max_radius=np.ceil(np.min(hists[0].shape)/4).astype(int)
    else:
        max_radius = np.ceil(max_radius/resolution).astype(int)
    if (linear_steps is None) or (linear_steps>=max_radius):
        linear_steps=min(max_radius,20)

    kernels,radii = _get_kernels(max_radius,linear_steps)

    co_occurrences = np.zeros((len(hists),)*2+(len(kernels),))

    shape = [hists[0].shape[a]+kernels[0].shape[a]-1 for a in [0,1]]
    fshape = [sp_fft.next_fast_len(shape[a], True) for a in [0,1]]

    kernels_fft = (sp_fft.rfftn(kernels, fshape,axes=[1,2]))

    width_kernel=kernels[0].shape[0]

    for i in range(len(hists)):
        print(i)
        h1_fft = sp_fft.rfftn(hists[i], fshape,axes=[0,1])
        h1_fftprod =  (h1_fft*kernels_fft)
        h1_conv = sp_fft.irfftn(h1_fftprod,fshape,axes=[1,2])
        h1_conv_ =  h1_conv[:,width_kernel//2:width_kernel//2+hists[0].shape[0],
                        width_kernel//2:width_kernel//2+hists[0].shape[1]] #signal._signaltools._centered(h1_conv,[len(kernels)]+fshape).copy()

        h1_product=h1_conv_*hists[i]/np.sum(kernels,axis=(1,2))[:,None,None]
        co_occurrences[i,i]=h1_product.sum(axis=(1,2))

        for j in range(i+1,len(hists)):
            h2_product=h1_conv_*hists[j]/np.sum(kernels,axis=(1,2))[:,None,None]
            co_occurrences[i,j] = h2_product.sum(axis=(1,2))
            co_occurrences[j,i]= co_occurrences[i,j]

    return _interpolate(radii, co_occurrences,resolution)
    # return radii, co_occurrences


def co_cooccurrence(sdata_0, sdata_1, resolution=5.0, max_radius=400, linear_steps=5, category=None):
    """co_occurrence _summary_

    :param sdata: SpatialData - all spots contained in sdata will be analysed. 
    :type sdata: SpatialData
    :param resolution: Smallest resolution of the co-occurence model in um, defaults to 5.
    :type resolution: float, optional
    :param max_radius: Largest radius of the co-occurrence model in um, defaults to 400.
    :type max_radius: float, optional
    :param linear_steps: Number of linear steps to model local heterogeniety. Afterwards, distance bins get wider to save computational resources, defaults to 5
    :type linear_steps: int, optional
    :param category: Category to model the co-occurrence of. Must be a column in 'sdata' with dtype 'category'. When given 'None', the algorithm defaults to gene labels.
    :type category: str, optional
    :return: _description_
    :rtype: _type_
    """

    mins = np.min([sdata_0.coordinates.min(
        0), sdata_1.coordinates.min(0)], axis=0)
    maxs = np.max([sdata_0.coordinates.max(
        0), sdata_1.coordinates.max(0)], axis=0)

    hists_0 = get_histograms(sdata_0, mins=mins, maxs=maxs, category=category)
    hists_1 = get_histograms(sdata_1, mins=mins, maxs=maxs, category=category)

    # return hists_0, hists_1

    if max_radius is None:
        max_radius = np.ceil(np.min(hists_0[0].shape)/4).astype(int)
    else:
        max_radius = np.ceil(max_radius/resolution).astype(int)
    if (linear_steps is None) or (linear_steps >= max_radius):
        linear_steps = min(max_radius, 20)

    kernels, radii = _get_kernels(max_radius, linear_steps)

    co_occurrences = np.zeros((len(hists_0), len(hists_1), len(kernels),))

    shape = [hists_0[0].shape[a]+kernels[0].shape[a]-1 for a in [0, 1]]
    fshape = [sp_fft.next_fast_len(shape[a], True) for a in [0, 1]]

    kernels_fft = (sp_fft.rfftn(kernels, fshape, axes=[1, 2]))

    width_kernel = kernels[0].shape[0]

    for i in range(len(hists_0)):
        print(i)
        h1_fft = sp_fft.rfftn(hists_0[i], fshape, axes=[0, 1])
        h1_fftprod = (h1_fft*kernels_fft)
        h1_conv = sp_fft.irfftn(h1_fftprod, fshape, axes=[1, 2])
        h1_conv_ = h1_conv[:, width_kernel//2:width_kernel//2+hists_0[0].shape[0],
                           width_kernel//2:width_kernel//2+hists_0[0].shape[1]]  # signal._signaltools._centered(h1_conv,[len(kernels)]+fshape).copy()

        # h1_product = h1_conv_*hists_0[i] / \
        #     np.sum(kernels, axis=(1, 2))[:, None, None]
        # co_occurrences[i, i] = h1_product.sum(axis=(1, 2))

        for j in range(i, len(hists_1)):
            h2_product = h1_conv_*hists_1[j] / \
                np.sum(kernels, axis=(1, 2))[:, None, None]
            co_occurrences[i, j] = h2_product.sum(axis=(1, 2))
            # co_occurrences[j, i] = co_occurrences[i, j]

    return _interpolate(radii, co_occurrences, resolution)
    # return radii, co_occurrences


def ripleys_k(sdata, resolution=5, max_radius=400, linear_steps=5, category=None):

    hists = get_histograms(sdata, category=category)  

    if max_radius is None:
        max_radius=np.ceil(np.min(hists[0].shape)/4).astype(int)
    else:
        max_radius = np.ceil(max_radius/resolution).astype(int)
    if (linear_steps is None) or (linear_steps>=max_radius):
        linear_steps=min(max_radius,20)

    kernels,radii = _get_kernels(max_radius,linear_steps)

    co_occurrences = np.zeros((len(hists),)*2+(len(kernels),))

    shape = [hists[0].shape[a]+kernels[0].shape[a]-1 for a in [0,1]]
    fshape = [sp_fft.next_fast_len(shape[a], True) for a in [0,1]]

    kernels_fft = (sp_fft.rfftn(kernels, fshape,axes=[1,2]))

    width_kernel=kernels[0].shape[0]

    for i in range(len(hists)):
        print(i)
        h1_fft = sp_fft.rfftn(hists[i], fshape,axes=[0,1])
        h1_fftprod =  (h1_fft*kernels_fft)
        h1_conv = sp_fft.irfftn(h1_fftprod,fshape,axes=[1,2])
        h1_conv_ =  h1_conv[:,width_kernel//2:width_kernel//2+hists[0].shape[0],
                        width_kernel//2:width_kernel//2+hists[0].shape[1]] #signal._signaltools._centered(h1_conv,[len(kernels)]+fshape).copy()

        h1_product=h1_conv_*hists[i]/np.sum(kernels,axis=(1,2))[:,None,None]
        co_occurrences[i,i]=h1_product.sum(axis=(1,2))

        for j in range(i+1,len(hists)):
            h2_product=h1_conv_*hists[j]/np.sum(kernels,axis=(1,2))[:,None,None]
            co_occurrences[i,j] = h2_product.sum(axis=(1,2))
            co_occurrences[j,i]= co_occurrences[i,j]

    return radii, co_occurrences

def mor_normalize(stats1,stats2):
    """mor_normalize _summary_

    :param stats1: _description_
    :type stats1: _type_
    :param stats2: _description_
    :type stats2: _type_
    """

    gmean = (stats1.counts*stats2.counts)**0.5

    ref_ratios_1 = (stats1.counts/gmean).fillna(0)
    ref_ratios_2 = (stats2.counts/gmean).fillna(0)

    norm_factor_1  = np.median(ref_ratios_1)
    norm_factor_2  = np.median(ref_ratios_2)

    normalized_1 = stats1.counts/norm_factor_1
    normalized_2 = stats2.counts/norm_factor_2

    return normalized_1,normalized_2
    


from setuptools import setup, find_packages

# here = pathlib.Path(__file__).parent.resolve()
print('kewl!!!!!!!!!')
if __name__ == "__main__":
    setup()
# setup(
#     name='plankton',
#     version='0.1.0',
#     description='A point process analysis project.',
#     package_dir={'': 'plankton'},
#     packages=find_packages('plankton'),
#     python_requires=">3.7",
#     install_requires=['numpy',
#                       'anndata',
#                       'scanpy',
#                       'jupyter',
#                       'matplotlib',
#                       'numpy',
#                       'pandas',
#                       'plotly',
#                       'scipy',
#                       'scikit-image',
#                       'scikit-learn',
#                       'umap-learn',
#                       ]
# )




# from __future__ import annotations
from tokenize import group

from matplotlib import cm
# from yaml import scan
# from sympy import N
from plankton.graph import SpatialGraph
from plankton.pixelmaps import PixelMap, PixelMask
import numpy as np
import pandas as pd
import random
import scipy
import anndata
from plankton.stats import co_occurrence
import scanpy as sc
import collections
from scipy import sparse
import pickle
from scipy.stats import binom

import matplotlib
from matplotlib.cm import get_cmap
from matplotlib.patches import ConnectionPatch
import matplotlib.patheffects as PathEffects
from matplotlib import pyplot as plt

import plotly.express as px
import plotly.graph_objects as go
from ipywidgets import widgets, interactive, HBox, VBox, Output, Layout

import warnings
warnings.simplefilter(action='ignore', category=UserWarning)


# plt.style.use('dark_background')
matplotlib.rcParams['figure.figsize'] = (15, 15)


class ScanpyDataFrame():
    """A plankton.py wrapper for scanpy-derived annotated data frames that provides an interface with many of the planktonpy functionalities.

    :param sdata: SpatialData the spatial data object that this ScanpyDataFrame 
    :type sdata: SpatialData
    :param scanpy_ds: the scanpy AnnData object containing the data set
    :type scanpy_ds: :class: anndata.AnnData
    """

    def __init__(self, sdata, scanpy_ds):
        """__init__ initialization method.

        """
        self.sdata = sdata
        self.adata = scanpy_ds
        self.stats = ScStatistics(self)
        self.celltype_labels = None
        self.signature_matrix = None

    @property
    def shape(self):
        return self.adata.shape

    @property
    def obs(self):
        return self.adata.obs

    @property
    def var(self):
        return self.adata.var

    @property
    def obsm(self):
        return self.adata.obsm

    @property
    def X(self):
        return self.adata.X

    def generate_signatures(self, celltype_obs_marker='celltype'):
        """generate_signatures Generates a signature matrix of gene->celltype correlations.

        :param celltype_obs_marker: column given in the 'obs' data field denoting the cell type, defaults to 'celltype'
        :type celltype_obs_marker: str, optional
        :return: signature matrix
        :rtype: pandas.DataFrame
        """
        self.celltype_labels = np.unique(self.adata.obs[celltype_obs_marker])

        self.signature_matrix = np.zeros((
            len(self.celltype_labels),
            self.adata.shape[1],
        ))

        for i, label in enumerate(self.celltype_labels):
            self.signature_matrix[i] = np.array(
                self.adata[self.adata.obs[celltype_obs_marker] == label].X.sum(
                    0)).flatten()

        self.signature_matrix = self.signature_matrix - self.signature_matrix.mean(
            1)[:, None]
        self.signature_matrix = self.signature_matrix / self.signature_matrix.std(
            1)[:, None]

        self.signature_matrix = pd.DataFrame(
            self.signature_matrix, index=self.celltype_labels, columns=self.stats.index)

        return self.signature_matrix

    def synchronize(self):
        """synchronizes the ScanpyDataFrame with its SpatialData reference objects. Unused genes are removed, statistics updated and sorted.
        """

        joined_genes = (self.stats.genes.intersection(
            self.sdata.genes)).sort_values()

        # print(len(joined_genes))

        self.sdata.reset_index()
        self.adata = self.adata[:, joined_genes]
        self.stats = ScStatistics(self)

        pd.DataFrame.drop(self.sdata, index=list(
            self.sdata.index[~self.sdata.g.isin(joined_genes)]),
            inplace=True)

        self.sdata.stats = PointStatistics(self.sdata)

        self.sdata.graph = SpatialGraph(self.sdata)

    def score_affinity(self, labels_1, labels_2=None, scanpy_obs_label='celltype'):

        if labels_2 is None:
            labels_2 = (self.adata.obs[~self.adata.obs[scanpy_obs_label].isin(labels_1)])[
                scanpy_obs_label]

        mask_1 = self.adata.obs[scanpy_obs_label].isin(labels_1)
        mask_2 = self.adata.obs[scanpy_obs_label].isin(labels_2)

        samples_1 = self.adata[mask_1, ]
        samples_2 = self.adata[mask_2, ]

        counts_1 = np.array(samples_1.X.mean(0)).flatten()
        counts_2 = np.array(samples_2.X.mean(0)).flatten()

        return np.log((counts_1+0.1)/(counts_2+0.1))

    def _repr_html_(self):

        magma = get_cmap('Greys')

        html = """  <style> 
                    .column-header{font-weight:bold;writing-mode:vertical-rl;text-orientation:mixed;}
                    .row-header{font-weight:bold;}
                    .x-value{border:collapse}
                    </style> <table>"""

        html += "<tr><td></td>"
        for i, h in enumerate(self.obs.index.values[:10]):
            html += f'<td><div class="column-header">{h}</div></td>'

        html += '<td>...</td>'

        for i, h in enumerate(self.obs.index.values[-10:]):
            html += f'<td><div class="column-header">{h}</div></td>'

        for i, k in enumerate(self.obs.keys()):
            html += "<tr>"
            html += f"<td>{k}</td>"
            for j, v in enumerate(self.obs[k].values[:10]):
                html += f"<td><div style='width:40px;overflow-wrap: break-word; '>{v}</div></td>"
            html += "<td>...</td>"

            for j, v in enumerate(self.obs[k].values[-10:]):
                html += f"<td><div style='width:40px;overflow-wrap: break-word; '>{v}</div></td>"
            html += "</tr>"

        html += "</tr>"

        X = self.X
        X_norm = X-X.min()
        X_norm /= X_norm.max()
        X_norm = X_norm.astype(float)
        X_norm = (X_norm)

        if type(X) == np.ndarray:
            def proc(x):
                return x
        else:
            def proc(x):
                return x.toarray().flatten()

        def return_cell(x):
            color = 'rgba' + \
                ''.join(str(magma(float(x)**0.8, bytes=True)).split(' '))
            return f"""<td bgcolor="{color}"><div class='x-value' style='width:100%;height:100%;color:white;font-weight:bold;'>{x:.2f}</div></td>"""

        for i, v in enumerate(self.var.index.values[:10]):
            html += f"<tr><td><div class='row-header'>{v}</div>"
            html += ''.join(list(map(return_cell, list(proc(X_norm[:10, i])))))
            if X_norm.shape[0] > 20:
                html += """<td bgcolor="white"><div class='x-value' style='width:100%;height:100%;'> ...</div></td>"""
            # html+=f"<tr><td><div class='row-header'>{v}</div>"
            html += ''.join(list(map(return_cell,
                            list(proc(X_norm[-10:, i])))))
            html += "</tr>"
        html += "<tr>" + """<td bgcolor="white"><div class='x-value' style='width:100%;height:100%;'> ...</div></td>"""*21+"</tr>"

        for i in range(-10, -1):
            html += f"<tr><td><div class='row-header'>{self.var.index.values[i]}</div>"
            html += ''.join(list(map(return_cell, list(proc(X_norm[:10, i])))))
            if X_norm.shape[0] > 20:
                html += """<td bgcolor="white"><div class='x-value' style='width:100%;height:100%;'> ...</div></td>"""
            html += ''.join(list(map(return_cell,
                            list(proc(X_norm[-10:, i])))))
            html += "</tr>"

        return html+"</table>"


class GeneStatistics(pd.DataFrame):
    """GeneStatistics Base class for the different stats data frames containing gene count data.

    """

    def __init__(self, *args, **kwargs):
        super(GeneStatistics, self).__init__(*args, **kwargs)

    @property
    def counts_sorted(self):
        return self.data.counts[self.stats.count_indices]

    @property
    def genes(self):
        return self.index

    def get_count(self, gene):
        """get_count returns the count for a given gene.

        :param gene: gene name
        :type gene: str
        :return: count value
        :rtype: int
        """
        if gene in self.genes.values:
            return int(self.counts[self.genes == gene])

    def get_id(self, gene_name):
        """get_id gets the id value for a certain gene

        :param gene_name: gene name
        :type gene_name: str
        :return: id value
        :rtype: int
        """
        return int(self.gene_ids[self.genes == gene_name])

    def get_count_rank(self, gene):
        if gene in self.genes.values:
            return int(self.count_ranks[self.genes == gene])

    def syncronized(self, stats):
        """syncronized returns a synchronized sdata count list for further statistical processing.
        NEEDS UPDATING!

        :param stats: external stats object to compare against
        :type stats: :class: plankton.GeneStatistics
        :return: concatenated count data frame
        :rtype: :class: pandas.DataFrame
        """
        return pd.concat([self, stats], axis=1).fillna(0)


class PointStatistics(GeneStatistics):
    """PointStatistics _summary_

    :param sdata: :class: SpatialData
    :type GeneStatistics: _type_
    """

    def __init__(self, sdata):

        counts = sdata.g.value_counts().sort_index()

        count_idcs = np.argsort(counts)
        count_ranks = np.argsort(count_idcs)

        super(PointStatistics, self).__init__(
            {
                'counts': counts,
                'count_ranks': count_ranks,
                'count_indices': count_idcs,
                'gene_ids': np.arange(len(counts))
            },
            index=counts.index)

        sdata['gene_id'] = (pd.Series(np.arange(len(counts)),
                            index=counts.index)[sdata.g]).values
        self.sdata = sdata

        sdata.graph = SpatialGraph(self)

    def co_occurrence(self, resolution=5, max_radius=None, linear_steps=20, category=None,):
        """co_occurrence Generate a co-occurrence matrix for this data set's source data.

        :param resolution: Smallest resolution of the co-occurence model in um, defaults to 5.
        :type resolution: float, optional
        :param max_radius: Largest radius of the co-occurrence model in um, defaults to 400.
        :type max_radius: float, optional
        :param linear_steps: Number of linear steps to model local heterogeniety. Afterwards, distance bins get wider to save computational resources, defaults to 5
        :type linear_steps: int, optional
        :param category: Category to model the co-occurrence of. Must be a column in 'sdata' with dtype 'category'. When given 'None', the algorithm defaults to gene labels.
        :rtype: str
        :return: co-occurrence matrix 
        :rtype: pandas.core.frame.DataFrame
        """
        return co_occurrence(self.sdata, resolution=resolution, max_radius=max_radius, linear_steps=linear_steps, category=category)

    def progressive_sample(self, power=1.2):
        count_ratios = (self.counts**(1/power)/self.counts)

        sample_mask = np.ones((len(self.sdata,)), dtype=bool)

        for i, c in enumerate(count_ratios):
            sample_array = np.linspace(0, 1, self.counts[i])
            np.random.shuffle(sample_array)
            sample_mask[self.sdata.gene_ids == i] = sample_array < c
        return sample_mask


class ScStatistics(GeneStatistics):
    """ScStatistics Statistics object for ScanpyDataFrame

    :param sdata: Source data frame 
    :type sdata: plankton.ScanpyDataFrame
    """

    def __init__(self, sdata):

        counts = np.array(sdata.adata.X.sum(0)).squeeze()
        genes = sdata.adata.var.index

        count_idcs = np.argsort(counts)
        count_ranks = np.argsort(count_idcs)

        super(ScStatistics, self).__init__(
            {
                'counts': counts,
                'count_ranks': count_ranks,
                'count_indices': count_idcs,
                'gene_ids': np.arange(len(genes))
            },
            index=genes)


class SpatialIndexer():
    """SpatialIndexer implements spatial view of a source plankton.SpatialData frame. It can be used for simple cropping and to extract spatial size information.

    :param sdata: source data
    :type sdata: plankton.SpatialData
    """

    def __init__(self, sdata):
        self.sdata = sdata

    @property
    def shape(self):
        """shape returns the shape of the data (instead of the dataframe shape). When available, the function reports the size of a (cropped) dataset background map. Otherwise, the extreme values in the x- and y- directions are returned.

        :return: Spatial extent/shape of the data.
        :rtype: tuple(int,int)
        """
        if self.sdata.background is None:
            return np.ceil(self.sdata.x.max() - self.sdata.x.min()).astype(
                int), np.ceil(self.sdata.y.max() - self.sdata.y.min()).astype(int)
        else:
            return self.sdata.background.shape

    def create_cropping_mask(self, start, stop, series):

        if start is None:
            start = 0

        if stop is None:
            stop = series.max()

        return ((series > start) & (series < stop))

    def join_cropping_mask(self, xlims, ylims):
        return self.create_cropping_mask(
            *xlims, self.sdata.x) & self.create_cropping_mask(*ylims, self.sdata.y)

    def crop(self, xlims, ylims):
        """crop crops the source data accordng to its spatial extent:

        :param xlims: limits in the x-direction
        :type xlims: (float,float)
        :param ylims: limits in the y-direction
        :type ylims: (float,float)
        :return: New, cropped data set.
        :rtype: plankton.SpatialData
        """

        mask = self.join_cropping_mask(xlims, ylims)

        pixel_maps = {}

        if xlims[0] is None:
            start_x = 0
        else:
            start_x = xlims[0]
        if ylims[0] is None:
            start_y = 0
        else:
            start_y = ylims[0]

        sdata = self.sdata[mask]
        sdata['x'] -= start_x
        sdata['y'] -= start_y

        for key, pm in sdata.pixel_maps.items():
            sdata.pixel_maps[key] = (pm[xlims[0]:xlims[1], ylims[0]:ylims[1]])

        return sdata

    def __getitem__(self, indices):
        """__getitem__ cropping interface for the source data set.

        :param indices: cropping indices.
        :type indices:  slice, iterable
        :return: cropped source data
        :rtype: plankton.SpatialData
        """

        if not isinstance(indices, collections.abc.Iterable):
            indices = (indices, )
        if len(indices) == 1:
            ylims = (0, None)
        else:
            ylims = (indices[1].start, indices[1].stop)

        # print(indices)

        xlims = (indices[0].start, indices[0].stop)

        return self.crop(xlims, ylims)


class VarDF(pd.DataFrame):
    """VarDF A dataframe encoding gene-centric observations.

    :param sdata: Source data
    :type sdata: plankton.SpatialData
    :param assign_colors: initialize a column with color-values for all genes, defaults to True
    :type assign_colors: bool, optional    
    """

    def __init__(self, sdata, assign_colors=True):

        super(VarDF, self).__init__(index=sdata.stats.index)
        self['genes'] = self.index
        if assign_colors:
            self.assign_colors()
        self.sdata = sdata

    def assign_colors(self, label='genes', cmap=None, shuffle=False):
        """assign_colors generates colors for a given label, which are stored in a column 'c_'+label

        :param label: column label to generate colors for, defaults to 'genes'
        :type label: str, optional
        :param cmap: colormap to pick colors from, defaults to None/'nipy_spectral'
        :type cmap: _type_, optional
        :param shuffle: Shuffle the colors randomly before assignment, defaults to False
        :type shuffle: bool, optional
        """

        uniques = self[label].unique()

        if cmap is None:
            cmap = get_cmap('nipy_spectral')
            clrs = [cmap(f) for f in np.linspace(0.07, 1, len(uniques))]

        else:
            cmap = get_cmap(cmap)
            clrs = [cmap(f) for f in np.linspace(0, 1, len(uniques))]

        if shuffle:
            random.shuffle(clrs)

        clrs = {u: clrs[i] for i, u in enumerate(uniques)}
        self['c_'+label] = self[label].astype('str').apply(lambda x: (clrs[x]))

    def project(self, label):
        """project Projects a column onto the gene composition of the source data set.

        :param label: Label of column to project.
        :type label: str
        """
        return(self.loc[self.sdata.g][label])

    def copy(self):
        return VarDF(self.sdata)


class SpatialData(pd.DataFrame):
    """SpatialData A data handling interface for Spot-Based spatial data.

    :param genes: A list of gene indicators
    :type genes: list(str)
    :param x_coordinates: a list of x-coordinates in micrometers.
    :type x_coordinates: list(float)
    :param y_coordinates: a list of y-coordinates in micrometers
    :type y_coordinates: list(float)
    :param pixel_maps: A dictionary of pixelmaps, can be used to define background images during initialization. defaults to {}
    :type pixel_maps: dict(str,plankton.pixelmaps.PixelMap), optional
    :param scanpy: An optional input for single-cell derived expression data. defaults to None
    :type scanpy: plankton.ScanpyDataFrame, optional
    :param synchronize: Whether the different data modalities should be cropped and synchronized at all times. defaults to True
    :type synchronize: bool, optional
    :param var: A data frame for gene-centric information, defaults to None
    :type var: plankton.VarDF, optional
    """

    def __init__(self,
                 genes,
                 x_coordinates,
                 y_coordinates,
                 pixel_maps={},
                 scanpy=None,
                 synchronize=True,
                 var=None):

        # Initiate 'own' spot data:
        super(SpatialData, self).__init__({
            'g': genes,
            'x': x_coordinates,
            'y': y_coordinates
        })

        self['g'] = self['g'].astype('category')

        self.stats = PointStatistics(self)

        if var is None:
            self.var = VarDF(self)
        else:
            self.var = var

        self.graph = SpatialGraph(self)

        # Initiate pixel maps:
        self.pixel_maps = {}
        self.bg_key = None

        for key, pm in pixel_maps.items():
            if self.bg_key is None:
                self.bg_key = key
            if isinstance(pm, np.ndarray):
                self.pixel_maps[key] = (PixelMap(pm))
            else:
                self.pixel_maps[key] = (pm)

        self.synchronize = synchronize

        # Append scanpy data set, synchronize both:
        if scanpy is not None:
            if type(scanpy) == ScanpyDataFrame:
                self.scanpy = self.scanpy = scanpy
            elif type(scanpy) == anndata.AnnData:
                self.scanpy = ScanpyDataFrame(self, scanpy)
            if self.synchronize:
                self.sync_scanpy()
        else:
            self.scanpy = None

        self.obsm = {"spatial":np.array(self.coordinates)}
        # self.obs = pd.DataFrame({'gene':self.g})
        self.uns = {}

        self.obsp = {}
        # self.obsm = {}

        self.varp = {}
        self.varm = {}

    @property
    def gene_ids(self):
        return self.gene_id

    @property
    def coordinates(self):
        return np.array([self.x, self.y]).T

    @property
    def counts(self):
        return self.stats['counts']

    @property
    def counts_sorted(self):
        return self.stats.counts[self.stats.count_indices]

    @property
    def genes_sorted(self):
        return self.genes[self.stats.count_indices]

    @property
    def genes(self):
        return self.stats.index

    @property
    def spatial(self):
        return SpatialIndexer(self)

    @property
    def background(self):
        if len(self.pixel_maps):
            return self.pixel_maps[self.bg_key]

    @property
    def adata(self):
        if self.scanpy is not None:
            return self.scanpy.adata

    @property
    def X(self):
        return scipy.sparse.csc_matrix((np.ones(len(self.g),), (np.arange(len(self.g)), np.array(self.gene_ids).flatten())),
                                       shape=(len(self.g), self.genes.shape[0],))

    @property
    def var_names(self):
        return self.var.columns
    # @property
    # def var(self):
    #     return pd.DataFrame(index=self.stats.genes)

    @property
    def obs(self):
        # pd.DataFrame({'gene': self.g}).astype(str).astype('category')
        return self

    # @property
    # def obsm(self):
    #     return {"spatial": self.coordinates}

    def __getitem__(self, *arg):

        if (len(arg) == 1):

            if type(arg[0]) == str:

                return super().__getitem__(arg[0])

            if (type(arg[0]) == slice):
                new_data = super().iloc.__getitem__(arg)

            elif (type(arg[0]) == int):
                new_data = super().iloc.__getitem__(slice(arg[0], arg[0] + 1))

            elif isinstance(arg[0], pd.Series):
                # print(arg[0].values)
                new_data = super().iloc.__getitem__(arg[0].values)

            elif isinstance(arg[0], np.ndarray):
                new_data = super().iloc.__getitem__(arg[0])

            elif isinstance(arg[0], collections.Sequence):
                if all([a in self.keys() for a in arg[0]]):
                    return super().__getitem__(*arg)
                new_data = super().iloc.__getitem__(arg[0])

            if self.scanpy is not None:
                scanpy = self.scanpy.adata
                synchronize = self.scanpy.synchronize
            else:
                scanpy = None
                synchronize = None

            new_frame = SpatialData(new_data.g,
                                    new_data.x,
                                    new_data.y,
                                    self.pixel_maps,
                                    scanpy=scanpy,
                                    synchronize=synchronize,
                                    var=self.var.copy())

            new_prop_entries = self.var.loc[new_frame.genes]
            new_frame.var[new_prop_entries.columns] = new_prop_entries
            new_frame.var.sdata = new_frame
            new_frame.var.drop(self.genes.symmetric_difference(
                new_frame.genes), inplace=True)

            for c in self.columns[4:]:
                new_frame[c] = new_data[c]

            if self.graph._umap is not None:
                new_frame.graph._umap = self.graph._umap[self.index.isin(
                    new_frame.index)]

            return (new_frame)

        print('Reverting to generic Pandas.')
        return super().__getitem__(*arg)

    def raw(self):
        return SpatialData(x_coordinates=self.x,
                           y_coordinates=self.y,
                           genes=self.g)

    def sync_scanpy(self,
                    mRNA_threshold_sc=1,
                    mRNA_threshold_spatial=1,
                    verbose=False,
                    anndata=None):
        """sync_scanpy Synchronizes the SpatialData set with its provided single-cell expression frame. Unpaired genes are removed and statistics re-calculated.

        :param mRNA_threshold_sc: Count below which a gene in the single-cell data set is removed, defaults to 1
        :type mRNA_threshold_sc: int, optional
        :param mRNA_threshold_spatial: Count below which a gene in the spatial data set is removed, defaults to 1
        :type mRNA_threshold_spatial: int, optional
        :param verbose: Displat warnings (legacy), defaults to False
        :type verbose: bool, optional
        :param anndata: _description_, defaults to None
        :type anndata: Provide a ScanpyDataFrame if the field 'self.scanpy' is empty, optional
        """
        if anndata is None and self.scanpy is None:
            print('Please provide some scanpy data...')

        if anndata is not None:
            self.scanpy = ScanpyDataFrame(anndata)
        else:
            self.scanpy.synchronize()

    def get_id(self, gene_name):
        """get_id Gets the dataframe-id to a certain gene

        :param gene_name: Gene name
        :type gene_name: str
        :return: Gene's id number that's shared across data sets/data modalities
        :rtype: int
        """
        return int(self.stats.gene_ids[self.genes == gene_name])

    def unite_columns(self, col_names, other='other'):
        """unite_columns Unites boolean columns in the data frame using their column names

        :param col_names: List of column names to unite
        :type col_names: iterable
        :param other: name for empty data field, defaults to 'other'
        :type other: str, optional
        """

        united = pd.Categorical(
            [other]*len(self), categories=[other]+list(col_names))
        for name in col_names:
            united[self[name]] = name

        return united

    def clean(self):

        new_frame = SpatialData(self.g.astype(str),
                                self.x,
                                self.y,
                                self.pixel_maps,
                                scanpy=self.scanpy,
                                synchronize=False,
                                var=self.var.copy())

        new_prop_entries = self.var.loc[new_frame.genes]
        new_frame.var[new_prop_entries.columns] = new_prop_entries
        new_frame.var.sdata = new_frame
        new_frame.var.drop(self.genes.symmetric_difference(
            new_frame.genes), inplace=True)

        for c in self.columns[4:]:
            new_frame[c] = self[c]

        if self.graph._umap is not None:
            new_frame.graph._umap = self.graph._umap[self.index.isin(
                new_frame.index)]

        return (new_frame)

    def drop(self, columns, inplace=False):
        """drop a column/list of columns

        :param columns: Column name or list of column names
        :type columns: str, iterable
        :param inplace: Whether to mutate the data in place or return a new object, defaults to False
        :type inplace: bool, optional
        :return: None if inplace=True, cropped SpatialData object otherwise. 
        :rtype: None, SpatialData
        """

        if inplace:
            # self.drop(column,inplace=True,axis=1)
            pd.DataFrame.drop(self, columns=columns, inplace=True)
        else:
            copy = self[:]
            # copy.drop(column,inplace=True,axis=1)
            pd.DataFrame.drop(copy, columns=columns, inplace=True)
            return copy

    def itergenes(self):
        return enumerate(self.genes)

    def scatter(self,
                c=None,
                color=None,
                color_prop='genes',
                marker='.',
                legend=None,
                axd=None,
                plot_bg=True,
                cmap='jet',
                scalebar=True,
                **kwargs):
        """scatter Leightweight implementation of a scatter plot, using maplotlib as a backend.

        :param c: Defines an individual color for each data point, either through a list of float scale values (color is sampled for a colormap) or a list of matplotlib-interpretable colors. defaults to None
        :type c: list(float),list(rgb), optional
        :param color: Marker color, defaults to None
        :type color: str, optional
        :param color_prop: column name in self.var, used as a color source, defaults to 'genes'
        :type color_prop: str, optional
        :param marker: Marker shape, defaults to '.'
        :type marker: str, optional
        :param legend: Whether to add a legend artist, defaults to None
        :type legend: bool, optional
        :param axd: Provide an axis object that the artists are added to. If none is provided, the current active axis is used. Defaults to None
        :type axd: matplotlib.pyplot.Asix, optional
        :param plot_bg: Whether to plot the background image, defaults to True
        :type plot_bg: bool, optional
        :param cmap: Define the colormap that colors are sampled from if a 'c' argument is provided, defaults to 'jet'
        :type cmap: str, optional
        :param scalebar: Whether to print a scalebar in micrometers, defaults to True
        :type scalebar: bool, optional
        :return: handles for the scatters,background and legend.
        :rtype: tuple
        """

        if axd is None:
            axd = plt.gca()

        axd.set_aspect('equal', adjustable='box')

        handle_imshow = None
        handle_legend = None

        if self.background and plot_bg:
            handle_imshow = self.background.imshow(
                cmap='Greys', axd=axd)
        else:
            axd.invert_yaxis()

        if c is None:
            if color is None:
                c = self.var.project('c_'+color_prop)

        if legend:
            # sorted(self.var[color_prop].unique())
            labels = self.g.unique().sort_values()
            # print(labels)
            clrs = [self.var[(self.var[color_prop] == l)]
                    ['c_'+color_prop][0] for l in labels]

            handles = [plt.scatter([], [], color=c) for c in clrs]
            handle_legend = plt.legend(handles, labels, loc='right')

        # axd.set_title(gene)
        handle_scatter = axd.scatter(self.x,
                                     self.y,
                                     c=c,
                                     marker=marker,
                                     color=color,
                                     cmap=cmap,
                                     **kwargs)

        if scalebar:
            self.add_scalebar(axd=axd)

        return handle_scatter, handle_imshow, handle_legend

    def add_scalebar(self, length=None, unit=r'$\mu$m', axd=None, color='w'):
        """add_scalebar: Adds a scalebar to a plot.

        :param length: Length of the scalebar, defaults to None
        :type length: float, optional
        :param unit: Unit that's displayed on the scalebar, defaults to r'$\mu'
        :type unit: str, optional
        :param axd: Axis to draw on, if none is proveded, the current active axis is used. Defaults to None
        :type axd:  matplotlib.pyplot.Axis, optional
        :param color:, Font/bar color. defaults to 'w'
        :type color: str, optional
        :return: handler for scalebar and text.
        :rtype: tuple
        """

        if axd is None:
            axd = plt.gca()

        x_, _x = axd.get_xlim()
        y_, _y = axd.get_ylim()

        if length is None:
            length = (_x-x_)*0.2
            decimals = np.ceil(np.log10(length))-1
            inter = length/10**decimals
            # int(np.around(_x-x_,1-int(np.log10((_x-x_)/2))))
            length = np.ceil(inter) * 10**decimals

        _x_ = _x-x_
        _y_ = _y-y_

        new_x_ = x_ + _x_/20
        new__x = new_x_+length
        new_y = y_ + _y_/20

        scbar = plt.Line2D([0.1, 0.9], [0.5, 0.5], c='w', marker='|', linewidth=2,
                           markeredgewidth=3, markersize=10, color=color)
        scbar.set_data([new_x_, new__x], [new_y, new_y])
        scbar.set_path_effects(
            [PathEffects.withStroke(linewidth=4, foreground='k')])

        magnitudes = ['nm', r'$\mu$m', 'mm', 'm', 'km']
        bar_label = f'{int(length%999)}{magnitudes[np.floor(np.log10(length)/3+1).astype(int)]}'

        sctxt = axd.text((new_x_+new__x)/2, (y_+_y_/15), bar_label,
                         fontweight='bold', ha='center', color=color)

        axd.add_artist(scbar)
        sctxt.set_path_effects(
            [PathEffects.withStroke(linewidth=3, foreground='k')])

        return scbar, sctxt

    def scatter_js(self, color_prop='c_genes'):
        """umap_js: A javascript-based display and selection function for the spatial source data and generated UMAP embedding. Can be used to understand UMAP agglomerations by displaying their gene compositions and projecting them back onto the original source data coordinates.

        :param color_prop: Property by which to color the spots in the scatter plots of the source data coordinates as well as the UMAP representation. Needs to be a column in sdata.var, defaults to 'c_genes'
        :type color_prop: str, optional
        :return: plotly HTML/javascript widget that displays in the jupyter notebook.
        :rtype: plotly.Widget
        """

        n_bars = 20

        trace_scatter = go.Scattergl(x=self.x,
                                     y=self.y,
                                     mode='markers',
                                     marker=dict(
                                         color=self.var.project(color_prop)),
                                     hoverinfo='none', meta={'name': 'tissue-scatter'},
                                     unselected={'marker': {
                                         'color': 'black', 'opacity': 0.2}},
                                     )

        f_scatter = go.FigureWidget(trace_scatter,)

        colors = self.var.loc[self.stats.sort_values(
            'counts')[-n_bars:].index, color_prop].values

        w_bars = go.Bar(x=self.stats.sort_values('counts')[-n_bars:].index,
                        y=self.stats.sort_values(
                            'counts')[-n_bars:]['counts'],
                        marker={
                            'color': ['rgb'+str(tuple((np.array(c)*256).astype(int))) for c in colors]},
                        )
        f_bars = go.FigureWidget(w_bars,
                                 layout={'title': 'n (selected):'},
                                 )

        f_bars.data[0]['showlegend'] = False

        w_bars_ratio_up = go.Bar(x=self.stats.sort_values('counts')[-n_bars:].index,
                                 y=[0]*n_bars,
                                 marker={
                                     'color': ['rgb'+str(tuple((np.array(c)*256).astype(int))) for c in colors]},
                                 )
        f_bars_ratio_up = go.FigureWidget(w_bars_ratio_up,
                                          layout={'title': 'n (selected) / total:'})

        f_bars_ratio_up.data[0]['showlegend'] = False

        w_bars_binom = go.Bar(x=self.stats.sort_values('counts')[-n_bars:].index,
                              y=[0]*n_bars,
                              marker={
                                  'color': ['rgb'+str(tuple((np.array(c)*256).astype(int))) for c in colors]},
                              )
        f_bars_binom = go.FigureWidget(w_bars_binom,
                                       layout={
                                           'title': 'binomcdf (n (selected),p(total)):'},
                                       )

        f_bars_binom.data[0]['showlegend'] = False

        out = widgets.Output(layout={'border': '1px solid black'})

        print('kewl')

        def update_bars(plot, points, selector):

            subset = self[np.array(points.point_inds)]

            colors = subset.var.loc[subset.stats.sort_values(
                'counts')[-n_bars:].index, color_prop].values
            colors = ['rgb'+str(tuple((np.array(c)*256).astype(int)))
                      for c in colors]
            ys = subset.stats.sort_values('counts')[-n_bars:]['counts']
            xs = subset.stats.sort_values('counts')[-n_bars:].index

            f_bars.data[0].marker.color = colors
            f_bars.data[0].x = xs
            f_bars.data[0].y = ys

            vals = (subset.stats.counts) / \
                (self.stats.loc[subset.stats.index].counts)
            idcs = np.argsort(vals)

            colors = subset.var[color_prop][idcs]
            colors = ['rgb'+str(tuple((np.array(c)*256).astype(int)))
                      for c in colors]
            ys = vals[idcs]
            xs = subset.stats.index[idcs]

            f_bars_ratio_up.data[0].marker.color = colors[-n_bars:]
            f_bars_ratio_up.data[0].x = xs[-n_bars:]
            f_bars_ratio_up.data[0].y = (ys[-n_bars:])

            vals = binom.cdf(subset.stats.counts, self.stats.loc[subset.stats.index].counts, len(
                subset)/len(self))
            idcs = np.argsort(vals)

            colors = subset.var[color_prop][idcs]
            colors = ['rgb'+str(tuple((np.array(c)*256).astype(int)))
                      for c in colors]
            ys = vals[idcs]
            xs = subset.stats.index[idcs]

            f_bars_binom.data[0].marker.color = colors[-n_bars:]
            f_bars_binom.data[0].x = xs[-n_bars:]
            f_bars_binom.data[0].y = (ys[-n_bars:])

        f_scatter.data[-1].on_selection(update_bars)
        # f_umap.data[0].on_selection(update_bars)

        text_field = widgets.Text(
            value='selection1',
            placeholder='Label for storing',
            description='Name:',
            disabled=False
        )

        store_button = widgets.Button(
            description='store selection',
            disabled=False,
            button_style='',  # 'success', 'info', 'warning', 'danger' or ''
        )

        def store_selection(event):
            self[text_field.value] = (pd.Series(
                np.arange(len(self))).isin(f_scatter.data[0].selectedpoints)).values

        def close_window(event):
            fig.data = []
            fig.layout = {}

        store_button.on_click(store_selection)

        reset_button = widgets.Button(
            description='close window',
            disabled=False,
            button_style='',  # 'success', 'info', 'warning', 'danger' or ''
        )

        reset_button.on_click(close_window)

        fig = widgets.VBox([widgets.HBox([f_scatter], layout=Layout(display='flex', width='100%', height='80%', border='red solid 1px', align_items='stretch', justify_content='space-around', flex_direction='row')),
                            widgets.HBox([widgets.HBox([f_bars, f_bars_ratio_up, f_bars_binom], layout=Layout(display='flex', width='80%')),
                                          widgets.VBox([widgets.HBox([text_field,
                                                                      ]), widgets.HBox([store_button, reset_button])], layout=Layout(display='flex', width='20%', height='100%', border='red solid 1px', justify_content='space-around', flex_direction='column')
                                                       )]
                                         )], layout=Layout(width='100%', height='80vh', background='red', border='solid 1px'))

        return fig

    def plot_bars(self, axis=None, **kwargs):
        """plot_bars: Plots a sorted bar  graph representation of the gene counts in the sample.

        :param axis: Axis to draw on, if none is proveded, the current active axis is used. Defaults to None
        :type axis: matplotlib.pyplot.Axis, optional
        """
        if axis is None:

            axis = plt.subplot(111)
        axis.bar(np.arange(len(self.stats.counts)), self.counts_sorted,
                 **kwargs)
        axis.set_yscale('log')

        axis.set_xticks(
            np.arange(len(self.genes_sorted)),
            list(self.genes_sorted),
        )
        axis.tick_params(axis='x', rotation=90)
        axis.set_ylabel('molecule count')

    def plot_overview(self):
        """plot_overview: Plots a lightweight overview of the spatial data set. A sorted bar representation of molecule counts is shown, alongside a few representative scatter plots of individual genes in the sample.
        """

        colors = ('royalblue', 'goldenrod', 'red', 'lime')

        scatter_idcs = np.round(np.linspace(0,
                                            len(self.stats.counts) - 1,
                                            4)).astype(int)

        fig, axd = plt.subplot_mosaic(
            [['scatter_1', 'scatter_2', 'scatter_3', 'scatter_4'],
             ['bar', 'bar', 'bar', 'bar']],
            figsize=(11, 7),
            constrained_layout=True)

        self.plot_bars(
            axd['bar'],
            color=[
                colors[np.where(
                    scatter_idcs == i)[0][0]] if i in scatter_idcs else 'grey'
                for i in range(len(self.stats.counts))
            ])

        for i in range(4):
            idx = self.stats.count_indices[scatter_idcs[i]]
            gene = self.genes[idx]
            plot_name = 'scatter_' + str(i + 1)
            axd[plot_name].set_title(gene)
            axd[plot_name].scatter(self.x, self.y, color=(0.5, 0.5, 0.5, 0.1))
            axd[plot_name].scatter(self.x[self['gene_id'] == idx],
                                   self.y[self['gene_id'] == idx],
                                   color=colors[i],
                                   marker='.')

            axd[plot_name].set_xticks([], [])
            # if i>0:
            axd[plot_name].set_yticks([], [])

            con = ConnectionPatch(xyA=(scatter_idcs[i],
                                       self.stats.counts[idx]),
                                  coordsA=axd['bar'].transData,
                                  xyB=(np.mean(axd[plot_name].get_xlim()),
                                       axd[plot_name].get_ylim()[0]),
                                  coordsB=axd[plot_name].transData,
                                  color='white',
                                  linewidth=1,
                                  linestyle='dotted')
            fig.add_artist(con)

        plt.suptitle('Selected Expression Densities:', fontsize=18)

    def squidpy(self, groupby=None):
        """squidpy: Generates an anndata.AnnData based squidpy-compatible object from the spatial data, including pixel maps and obervation matrices.

        :param groupby: Name of a dataset column (of type 'category') that denotes membership to certain spatial entities (such as cells) by which to agglomerate the data. I none is provided, every data point is grouped individually. Defaults to None
        :type groupby: str, optional
        :return: squidpy-compatible spatial data set. 
        :rtype: anndata.AnnData
        """
        # obs={"cluster":self.gene_id.astype('category')}

        if groupby is None:
            obsm = {"spatial": np.array(self.coordinates)}

            uns = self.uns.update({'Image': self.background})
            obs = pd.DataFrame({'gene': self.g})
            obs['gene'] = obs['gene'].astype('category')
            return anndata.AnnData(X=self.X, obs=obs, var=self.var, obsm=obsm)

        else:

            uns = self.uns.update({'Image': self.background})
            obs = pd.DataFrame({groupby: self[groupby].unique()})
            obsm = {"spatial": np.zeros((len(obs), 2))}
            var = self.var

            X = np.zeros((len(obs), len(var)))

            for i, o in enumerate(obs[groupby].values):
                subset = self[self[groupby] == o]
                obsm['spatial'][i] = subset.coordinates.mean(0)
                locs = [self.genes.get_loc(i) for i in subset.counts.index]
                X[i, locs] = subset.counts.values

            return anndata.AnnData(X=X, obs=obs, var=var, obsm=obsm)

    def save(self, path):
        """save: Save a deep copy of the spatial data, including graphs and pixelmaps.

        :param path: Save path/filename
        :type path: str
        """
        pickle.dump({'sdata': self, 'graph': self.graph,
                    'var': self.var, 'pixel_maps': self.pixel_maps}, open(path, "wb"))



def load(path):
    """load: Loads a spatial data set from a disk file.

    :param path: File path/name
    :type path: str
    :return: spatial data set
    :rtype: plankton.SpatialData
    """
    data = pickle.load(open(path, "rb"))
    sdata = SpatialData(data['sdata']['g'], data['sdata']
                        ['x'], data['sdata']['y'], pixel_maps=data['pixel_maps'])
    print(data['sdata'].columns)

    for i, c in enumerate(data['sdata'].columns[4:]):
        sdata[c] = data['sdata'][c]
    sdata.graph = data['graph']
    sdata.graph.sdata = sdata
    sdata.var = data['var']
    sdata.var.sdata = sdata
    return sdata

def load_legacy(path):
    """load: Loads a spatial data set from a disk file.

    :param path: File path/name
    :type path: str
    :return: spatial data set
    :rtype: plankton.SpatialData
    """
    data = pickle.load(open(path, "rb"))
    sdata = SpatialData(data['sdata']['g'], data['sdata']
                        ['x'], data['sdata']['y'], pixel_maps=data['pixel_maps'])
    print(data['sdata'].columns)

    for i, c in enumerate(data['sdata'].columns[4:]):
        sdata[c] = data['sdata'][c]
    sdata.graph = data['graph']
    sdata.graph.sdata = sdata

    if 'var' in data.keys():
        sdata.var=data['var']
    else:
        sdata.var=data['obsc']

    sdata.var.sdata = sdata
    return sdata



from importlib.resources import path
from textwrap import fill
from multiprocessing import Pool

import numpy as np
import community
import networkx as nx

# from requests import patch
from scipy.ndimage import gaussian_filter, maximum_filter
from plankton.pixelmaps import PixelMap

from sklearn.neighbors import NearestNeighbors,kneighbors_graph
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans,DBSCAN
from sklearn import preprocessing

from skimage import filters

from umap import UMAP

import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap

from tqdm import tqdm_notebook
import scipy
from scipy.spatial.distance import cdist


def fill_celltypemaps(ct_map, fill_blobs=True, min_blob_area=0, filter_params={}, output_mask=None):
    """
    Post-filter cell type maps created by `map_celltypes`.

    :param min_r: minimum threshold of the correlation.
    :type min_r: float
    :param min_norm: minimum threshold of the vector norm.
        If a string is given instead, then the threshold is automatically determined using
        sklearn's `threshold filter functions <https://scikit-image.org/docs/dev/api/skimage.filters.html>`_ (The functions start with `threshold_`).
    :type min_norm: str or float
    :param fill_blobs: If true, then the algorithm automatically fill holes in each blob.
    :type fill_blobs: bool
    :param min_blob_area: The blobs with its area less than this value will be removed.
    :type min_blob_area: int
    :param filter_params: Filter parameters used for the sklearn's threshold filter functions.
        Not used when `min_norm` is float.
    :type filter_params: dict
    :param output_mask: If given, the cell type maps will be filtered using the output mask.
    :type output_mask: np.ndarray(bool)
    """

    from skimage import measure

    filtered_ctmaps = np.zeros_like(ct_map) - 1

    for cidx in np.unique(ct_map):
        mask = ct_map == cidx
        if min_blob_area > 0 or fill_blobs:
            blob_labels = measure.label(mask, background=0)
            for bp in measure.regionprops(blob_labels):
                if min_blob_area > 0 and bp.filled_area < min_blob_area:
                    for c in bp.coords:
                        mask[c[0], c[1], ] = 0

                    continue
                if fill_blobs and bp.area != bp.filled_area:
                    minx, miny,  maxx, maxy, = bp.bbox
                    mask[minx:maxx, miny:maxy, ] |= bp.filled_image

        filtered_ctmaps[np.logical_and(mask == 1, np.logical_or(
            ct_map == -1, ct_map == cidx))] = cidx

    return filtered_ctmaps


def get_histograms(sdata, mins=None, maxs=None, category=None, resolution=5):

    if mins is None:
        mins = sdata.coordinates.min(0)

    if maxs is None:
        maxs = sdata.coordinates.max(0)+1

    # print(mins,maxs)

    n_bins = np.ceil(np.divide(maxs-mins, resolution)).astype(int)

    histograms = []

    if category is None:
        print(mins,maxs)
        for gene in sdata.genes:
            histograms.append(np.histogram2d(
                *sdata[sdata.g == gene].coordinates.T, bins=n_bins, range=([mins[0], maxs[0]], [mins[1], maxs[1]]))[0])

    else:
        for c in sdata[category].cat.categories:
            histograms.append(np.histogram2d(
                *sdata[sdata[category] == c].coordinates.T, bins=n_bins, range=([mins[0], maxs[0]], [mins[1], maxs[1]]))[0])

    return histograms,


def crosscorr(x, y):
    x -= np.array(x.mean(1))[:, None]
    y -= np.array(y.mean(1))[:, None]
    c = (np.dot(x, y.T)/x.shape[1]).squeeze()

    return np.nan_to_num(np.nan_to_num(c/np.array(x.std(1))[:, None])/np.array(y.std(1))[None, :])

def ssam(sdata, signatures=None, adata_obs_label='celltype', kernel_bandwidth=2.5, output_um_p_px=5,
         patch_length=1000, threshold_exp=0.1, threshold_cor=0.1, background_value=-1,
         fill_blobs=True, min_blob_area=10):

    if signatures is None:
        signatures = sdata.uns['ssam']['centroids']

    # if (sdata.scanpy is not None):
        # signatures = sdata.scanpy.generate_signatures(adata_obs_label)

    kernel_bandwidth_px = kernel_bandwidth/output_um_p_px

    out_shape = (np.ceil(sdata.x.max()/output_um_p_px+kernel_bandwidth_px*3).astype(int),
                 np.ceil(sdata.y.max()/output_um_p_px+kernel_bandwidth_px*3).astype(int))

    ct_map = np.zeros((out_shape), dtype=int)+background_value
    vf_norm = np.zeros_like(ct_map)

    range_x = np.floor(sdata.x.min()).astype(
        int), np.ceil(sdata.x.max()).astype(int)
    patch_delimiters_x = list(
        range(range_x[0], range_x[1], patch_length))+[range_x[1]]

    range_y = np.floor(sdata.y.min()).astype(
        int), np.ceil(sdata.y.max()).astype(int)
    patch_delimiters_y = list(
        range(range_y[0], range_y[1], patch_length))+[range_y[1]]

    print(list(patch_delimiters_x), list(patch_delimiters_y))

    with tqdm_notebook(total=(len(patch_delimiters_x)-1)*(len(patch_delimiters_y)-1)) as pbar:
        for i, x in enumerate(patch_delimiters_x[:-1]):
            for j, y in enumerate(patch_delimiters_y[:-1]):
                sdata_patch = sdata.raw().spatial[x:patch_delimiters_x[i+1],
                                                y:patch_delimiters_y[j+1]]

                if not len(sdata_patch):
                    break

                mins = sdata_patch.coordinates.min(0).astype(int)
                maxs = sdata_patch.coordinates.max(
                    0).astype(int)+kernel_bandwidth*3
                hists = get_histograms(
                    sdata_patch, mins=mins, maxs=maxs, resolution=output_um_p_px)
                hists = np.concatenate(
                    [gaussian_filter(h, kernel_bandwidth_px) for h in hists])

                # print(hists.shape)

                norm = np.sum(hists, axis=0)

                vf_norm[(x+mins[0])//output_um_p_px:(x+mins[0])//output_um_p_px+norm.shape[0],
                        (y+mins[1])//output_um_p_px:(y+mins[1])//output_um_p_px+norm.shape[1], ] = norm

                mask = norm > threshold_exp

                exps = np.zeros((len(sdata.genes), mask.sum()))

                # print(exps.shape,signatures.shape)
                exps[sdata.stats.loc[sdata_patch.genes].gene_ids.values,
                    :] = hists[:, mask]

                local_ct_map = mask.astype(int)-1

                corrs = crosscorr(exps.T, signatures)
                # print(corrs.min())
                corrs_winners = corrs.argmax(1)
                corrs_winners[corrs.max(1) < threshold_cor] = background_value
                local_ct_map[mask] = corrs_winners

                # idcs=sdata_patch.index

                x_ = int(max(0, (x+mins[0])-kernel_bandwidth_px*3))//output_um_p_px
                y_ = int(max(0, (y+mins[1])-kernel_bandwidth_px*3))//output_um_p_px

                ct_map[x_:x_+local_ct_map.shape[0],
                    y_:y_+local_ct_map.shape[1]] = local_ct_map

                pbar.update(1)


    if not 'ssam' in sdata.uns.keys():
        sdata.uns['ssam']={}

    sdata.uns['ssam']['ct_map_raw']=ct_map
    sdata.uns['ssam']['vf_norm']=vf_norm

    ct_map = fill_celltypemaps(
        ct_map, min_blob_area=min_blob_area, fill_blobs=fill_blobs)


    sdata.uns['ssam']['ct_map_filtered']=ct_map

    return PixelMap(ct_map.T, px_p_um=1/output_um_p_px)


def localmax_sampling(sdata, min_distance=3, bandwidth=4, min_signal_norm=0.2,n_neighbors=150):
    n_bins = np.array(sdata.coordinates.max(0)).astype(int)+2

    print(n_bins)
    vf = (gaussian_filter(np.histogram2d(
        *sdata.coordinates.T, bins=(range(n_bins[0]),range(n_bins[1])))[0], bandwidth))

    # return vf
    localmaxs = np.where((maximum_filter(vf, min_distance) == vf) & (vf > min_signal_norm))
    knn = NearestNeighbors(n_neighbors=n_neighbors)
    knn.fit(sdata.coordinates)
    dists, nbrs = knn.kneighbors(np.array(localmaxs).T)
    neighbor_types = np.array(sdata.gene_ids)[nbrs]

    localmax_samples = np.zeros((dists.shape[0], len(sdata.genes)))

    bandwidth = bandwidth
    def kernel(x): return np.exp(-x**2/(2*bandwidth**2))

    for i in range(0, dists.shape[1]):
        localmax_samples[np.arange(dists.shape[0]), neighbor_types[:, i]
               ] += kernel(dists[:, i])

    if not 'ssam' in sdata.uns.keys():
        sdata.uns['ssam']={}

    sdata.uns['ssam']['localmax_coords']=localmaxs
    sdata.uns['ssam']['localmax_samples']=localmax_samples

    return localmax_samples,localmaxs

def normalize_vectors(sdata, use_expanded_vectors=False, normalize_gene=False, normalize_vector=False, normalize_median=False, 
                        norm='l1',size_after_normalization=1e4, log_transform=False, scale=False):
    """
    Normalize and regularize vectors

    :param use_expanded_vectors: If True, use averaged vectors nearby local maxima of the vector field.
    :type use_expanded_vectors: bool
    :param normalize_gene: If True, normalize vectors by sum of each gene expression across all vectors.
    :type normalize_gene: bool
    :param normalize_vector: If True, normalize vectors by sum of all gene expression of each vector.
    :type normalize_vector: bool
    :param log_transform: If True, vectors are log transformed.
    :type log_transform: bool
    :param scale: If True, vectors are z-scaled (mean centered and scaled by stdev).
    :type scale: bool
    """
    if use_expanded_vectors:
        vec = np.array(sdata.uns['ssam']['expanded_vectors'], copy=True)
    else:
        vec = np.array(sdata.uns['ssam']['localmax_samples'], copy=True)
    if normalize_gene:
        vec = preprocessing.normalize(vec, norm=norm, axis=0) * size_after_normalization  # Normalize per gene
    if normalize_vector:
        vec = preprocessing.normalize(vec, norm="l1", axis=1) * size_after_normalization # Normalize per vector
    if normalize_median:
        def n(v):
            s, m = np.sum(v, axis=1), np.median(v, axis=1)
            s[m > 0] = s[m > 0] / m[m > 0]
            s[m == 0] = 0
            v[s > 0] = v[s > 0] / s[s > 0][:, np.newaxis]
            v[v == 0] = 0
            return v
        vec = n(vec)
    if log_transform:
        vec = np.log2(vec + 1)
    if scale:
        vec = preprocessing.scale(vec)
    sdata.uns['ssam']['normalized_vectors'] = vec
    return

def cluster_vectors(sdata, pca_dims=10, min_cluster_size=0, resolution=0.6, prune=1.0/15.0, snn_neighbors=30, max_correlation=1.0,
                    metric="correlation", subclustering=False, dbscan_eps=0.4, centroid_correction_threshold=0.8, random_state=0):
    """
    Cluster the given vectors using the specified clustering method.

    :param pca_dims: Number of principal componants used for clustering.
    :type pca_dims: int
    :param min_cluster_size: Set minimum cluster size.
    :type min_cluster_size: int
    :param resolution: Resolution for Louvain community detection.
    :type resolution: float
    :param prune: Threshold for Jaccard index (weight of SNN network). If it is smaller than prune, it is set to zero.
    :type prune: float
    :param snn_neighbors: Number of neighbors for SNN network.
    :type snn_neighbors: int
    :param max_correlation: Clusters with higher correlation to this value will be merged.
    :type max_correlation: bool
    :param metric: Metric for calculation of distance between vectors in gene expression space.
    :type metric: str
    :param subclustering: If True, each cluster will be clustered once again with DBSCAN algorithm to find more subclusters.
    :type subclustering: bool
    :param centroid_correction_threshold: Centroid will be recalculated with the vectors
        which have the correlation to the cluster medoid equal or higher than this value.
    :type centroid_correction_threshold: float
    :param random_state: Random seed or scikit-learn's random state object to replicate the same result
    :type random_state: int or random state object
    """
    
    vecs_normalized = sdata.uns['ssam']['normalized_vectors']
    vecs_normalized_dimreduced = PCA(n_components=pca_dims, random_state=random_state).fit_transform(vecs_normalized)

    print(vecs_normalized_dimreduced.shape)

    def cluster_vecs(vecs):
        k = min(snn_neighbors, vecs.shape[0])
        knn_graph = kneighbors_graph(vecs, k, mode='connectivity', include_self=True, metric=metric).todense()
        intersections = np.dot(knn_graph, knn_graph.T)
        snn_graph = intersections / (k + (k - intersections)) # borrowed from Seurat
        snn_graph[snn_graph < prune] = 0
        G = nx.from_numpy_matrix(snn_graph)
        partition = community.best_partition(G, resolution=resolution, random_state=random_state)
        lbls = np.array(list(partition.values()))
        return lbls

    def remove_small_clusters(lbls, lbls2=None):
        small_clusters = []
        cluster_indices = []
        lbls = np.array(lbls)
        for lbl in np.unique(lbls):
            if lbl == -1:
                continue
            cnt = np.sum(lbls == lbl)
            if cnt < min_cluster_size:
                small_clusters.append(lbl)
            else:
                cluster_indices.append(lbl)
        for lbl in small_clusters:
            lbls[lbls == lbl] = -1
        tmp = np.array(lbls, copy=True)
        for i, idx in enumerate(cluster_indices):
            lbls[tmp == idx] = i
        if lbls2 is not None:
            for lbl in small_clusters:
                lbls2[lbls2 == lbl] = -1
            tmp = np.array(lbls2, copy=True)
            for i, idx in enumerate(cluster_indices):
                lbls2[tmp == idx] = i
            return lbls, lbls2
        else:
            return lbls
    
    if subclustering:
        super_lbls = cluster_vecs(vecs_normalized_dimreduced)
        dbscan = DBSCAN(eps=dbscan_eps, min_samples=min_cluster_size, metric=metric)
        all_lbls = np.zeros_like(super_lbls)
        global_lbl_idx = 0
        for super_lbl in set(list(super_lbls)):
            super_lbl_idx = np.where(super_lbls == super_lbl)[0]
            if super_lbl == -1:
                all_lbls[super_lbl_idx] = -1
                continue
            sub_lbls = dbscan.fit(vecs_normalized_dimreduced[super_lbl_idx]).labels_
            for sub_lbl in set(list(sub_lbls)):
                if sub_lbl == -1:
                    all_lbls[tuple([super_lbl_idx[sub_lbls == sub_lbl]])] = -1
                    continue
                all_lbls[tuple([super_lbl_idx[sub_lbls == sub_lbl]])] = global_lbl_idx
                global_lbl_idx += 1
    else:
        all_lbls = cluster_vecs(vecs_normalized_dimreduced)            
            
    new_labels = __correct_cluster_labels(vecs_normalized,all_lbls, centroid_correction_threshold)
    new_labels, all_lbls = remove_small_clusters(new_labels, all_lbls)
    centroids, centroids_stdev = __calc_centroid(vecs_normalized, new_labels)

    merge_candidates = []
    if max_correlation < 1.0:
        Z = scipy.cluster.hierarchy.linkage(centroids, metric='correlation')
        clbls = scipy.cluster.hierarchy.fcluster(Z, 1 - max_correlation, 'distance')
        for i in set(clbls):
            leaf_indices = np.where(clbls == i)[0]
            if len(leaf_indices) > 1:
                merge_candidates.append(leaf_indices)
        removed_indices = []
        for cand in merge_candidates:
            for i in cand[1:]:
                all_lbls[all_lbls == i] = cand[0]
                removed_indices.append(i)
        for i in sorted(removed_indices, reverse=True):
            all_lbls[all_lbls > i] -= 1

        new_labels = __correct_cluster_labels(vecs_normalized,all_lbls, centroid_correction_threshold)
        new_labels, all_lbls = remove_small_clusters(new_labels, all_lbls)
        centroids, centroids_stdev = __calc_centroid(vecs_normalized, new_labels)
            
    sdata.uns['ssam']['cluster_labels'] = all_lbls
    sdata.uns['ssam']['filtered_cluster_labels'] = new_labels
    sdata.uns['ssam']['centroids'] = np.array(centroids)
    sdata.uns['ssam']['centroids_stdev'] = np.array(centroids_stdev)
    #self.dataset.medoids = np.array(medoids)
    
    print("Found %d clusters"%len(centroids))
    return

def __correct_cluster_labels(normalized_vectors, cluster_labels, centroid_correction_threshold):
    new_labels = np.array(cluster_labels, copy=True)
    if centroid_correction_threshold < 1.0:
        for cidx in np.unique(cluster_labels):
            if cidx == -1:
                continue
            prev_midx = -1
            while True:
                vecs = normalized_vectors[new_labels == cidx]
                vindices = np.where(new_labels == cidx)[0]
                midx = vindices[np.argmin(np.sum(cdist(vecs, vecs), axis=0))]
                if midx == prev_midx:
                    break
                prev_midx = midx
                m = normalized_vectors[midx]
                for vidx, v in zip(vindices, vecs):
                    if crosscorr(v[None], m[None]) < centroid_correction_threshold:
                        new_labels[vidx] = -1
    return new_labels

def __calc_centroid(normalized_vectors, cluster_labels):
    centroids = []
    centroids_stdev = []
    #medoids = []
    for lbl in sorted(list(set(cluster_labels))):
        if lbl == -1:
            continue
        cl_vecs = normalized_vectors[cluster_labels == lbl, :]
        #cl_dists = scipy.spatial.distance.cdist(cl_vecs, cl_vecs, metric)
        #medoid = cl_vecs[np.argmin(np.sum(cl_dists, axis=0))]
        centroid = np.mean(cl_vecs, axis=0)
        centroid_stdev = np.std(cl_vecs, axis=0)
        #medoids.append(medoid)
        centroids.append(centroid)
        centroids_stdev.append(centroid_stdev)
    return centroids, centroids_stdev#, medoids

def __run_pca(sdata, exclude_bad_clusters, pca_dims, random_state):
    if exclude_bad_clusters:
        good_vecs = sdata.uns['ssam']['normalized_vectors'][sdata.uns['ssam']['filtered_cluster_labels'] != -1, :]
    else:
        good_vecs = sdata.uns['ssam']['normalized_vectors']
    return PCA(n_components=pca_dims, random_state=random_state).fit_transform(good_vecs)

def generate_umap(sdata,run_umap=False, pca_dims=10, metric="correlation", exclude_bad_clusters=True, random_state=0, umap_kwargs={}):
    """generates UMAP of localax coordinates, stored in sdata.uns['ssam']['umap']"""

    if sdata.uns['ssam']['filtered_cluster_labels'] is None:
        exclude_bad_clusters = False
    if run_umap or not ('umap' in sdata.uns['ssam'].keys()):
        pcs = __run_pca(sdata,exclude_bad_clusters, pca_dims, random_state)
        sdata.uns['ssam']['umap'] = UMAP(metric=metric, random_state=random_state, **umap_kwargs).fit_transform(pcs[:, :pca_dims])

def plot_umap(sdata,colors=[],exclude_bad_clusters=True, s=None, excluded_color="#00000033", cmap="jet",):
    """Plots a umap of localmax expression values generated in the function 'generate_umap' """

    if sdata.uns['ssam']['filtered_cluster_labels'] is not None:
        cols = sdata.uns['ssam']['filtered_cluster_labels'][sdata.uns['ssam']['filtered_cluster_labels'] != -1]
    else:
        cols = None

    if len(colors) > 0:
        cmap = ListedColormap(colors)
    if not exclude_bad_clusters and sdata.uns['ssam']['filtered_cluster_labels'] is not None:
        plt.scatter(sdata.uns['ssam']['umap'][:, 0][sdata.uns['ssam']['filtered_cluster_labels'] == -1], sdata.uns['ssam']['umap'][:, 1][sdata.uns['ssam']['filtered_cluster_labels'] == -1], s=s, c=excluded_color)
        plt.scatter(sdata.uns['ssam']['umap'][:, 0][sdata.uns['ssam']['filtered_cluster_labels'] != -1], sdata.uns['ssam']['umap'][:, 1][sdata.uns['ssam']['filtered_cluster_labels'] != -1], s=s, c=cols, cmap=cmap)
    else:
        plt.scatter(sdata.uns['ssam']['umap'][:, 0], sdata.uns['ssam']['umap'][:, 1], s=s, c=cols, cmap=cmap)
    return


def plot_umap(sdata,colors=[],exclude_bad_clusters=True, s=None, excluded_color="#00000033", cmap="jet",):
    """Plots a umap of localmax expression values generated in the function 'generate_umap' """

    if sdata.uns['ssam']['filtered_cluster_labels'] is not None:
        cols = sdata.uns['ssam']['filtered_cluster_labels'][sdata.uns['ssam']['filtered_cluster_labels'] != -1]
    else:
        cols = None

    if len(colors) > 0:
        cmap = ListedColormap(colors)
    if not exclude_bad_clusters and sdata.uns['ssam']['filtered_cluster_labels'] is not None:
        plt.scatter(sdata.uns['ssam']['umap'][:, 0][sdata.uns['ssam']['filtered_cluster_labels'] == -1], sdata.uns['ssam']['umap'][:, 1][sdata.uns['ssam']['filtered_cluster_labels'] == -1], s=s, c=excluded_color)
        plt.scatter(sdata.uns['ssam']['umap'][:, 0][sdata.uns['ssam']['filtered_cluster_labels'] != -1], sdata.uns['ssam']['umap'][:, 1][sdata.uns['ssam']['filtered_cluster_labels'] != -1], s=s, c=cols, cmap=cmap)
    else:
        plt.scatter(sdata.uns['ssam']['umap'][:, 0], sdata.uns['ssam']['umap'][:, 1], s=s, c=cols, cmap=cmap)
    return


def plot_localmax(sdata, c=None, cmap=None, s=1, rotate=0):
    """
    Scatter plot the local maxima.

    :param c: Color of the scatter dots. Overrides `cmap` parameter.
    :type c: str or list(str), or list(float) or list(list(float))
    :param cmap: Colormap of the scatter dots.
    :type cmap: str or matplotlib.colors.Colormap
    :param s: Size of the scatter dots.
    :param rotate: Rotate the plot. Possible values are 0, 1, 2, and 3.
    :type rotate: int
    """
    if rotate < 0 or rotate > 3:
        raise ValueError("rotate can only be 0, 1, 2, 3")
    if rotate == 0 or rotate == 2:
        dim0, dim1 = 1, 0
    elif rotate == 1 or rotate == 3:
        dim0, dim1 = 0, 1
    plt.scatter(sdata.uns['ssam']['localmax_coords'][dim0], sdata.uns['ssam']['localmax_coords'][dim1], s=s, c=c, cmap=cmap)
    plt.xlim([0, sdata.uns['ssam']['vf_norm'].shape[dim0]])
    plt.ylim([sdata.uns['ssam']['vf_norm'].shape[dim1], 0])
    if rotate == 1:
        plt.gca().invert_xaxis()
    elif rotate == 2:
        plt.gca().invert_xaxis()
        plt.gca().invert_yaxis()
    elif rotate == 3:
        plt.gca().invert_yaxis()    


def plot_celltypes_map(sdata, background="black", centroid_indices=[], colors=None, cmap='jet', rotate=0, min_r=0.6, set_alpha=False):
    """
    Plot the merged cell-type map.

    :param background: Set background color of the cell-type map.
    :type background: str or list(float)
    :param centroid_indices: The centroids which will be in the cell type map. If not given, the cell-type map is drawn with all centroids.
    :type centroid_indices: list(int)
    :param colors: Color of the clusters. Overrides `cmap` parameter.
    :type colors: list(str), list(list(float))
    :param cmap: Colormap for the clusters.
    :type cmap: str or matplotlib.colors.Colormap
    :param rotate: Rotate the plot. Possible values are 0, 1, 2, and 3.
    :type rotate: int
    :param min_r: Minimum correlation threshold for the cell-type map.
        This value is only for the plotting, does not affect to the cell-type maps generated by `filter_celltypemaps`.
    :type min_r: float
    :param set_alpha: Set alpha of each pixel based on the correlation.
        Not properly implemented yet, doesn't work properly with the background other than black.
    :type set_alpha: bool
    :param z: Z index to slice 3D cell-type map.
        If not given, the slice at the middle will be used.
    :type z: int
    """
    # if z is None:
    #     z = int(sdata.uns['ssam']['filtered_'].shape[2] / 2)
    num_ctmaps = np.max(sdata.uns['ssam']['ct_map_filtered']) + 1
    
    if len(centroid_indices) == 0:
        centroid_indices = list(range(num_ctmaps))
        
    if colors is None:
        cmap_internal = plt.get_cmap(cmap)
        colors = cmap_internal([float(i) / (num_ctmaps - 1) for i in range(num_ctmaps)])
        
    all_colors = [background if not j in centroid_indices else colors[i] for i, j in enumerate(range(num_ctmaps))]
    cmap_internal = ListedColormap(all_colors)

    celltype_maps_internal = np.array(sdata.uns['ssam']['ct_map_filtered'], copy=True)
    empty_mask = celltype_maps_internal == -1
    celltype_maps_internal[empty_mask] = 0
    sctmap = cmap_internal(celltype_maps_internal)
    sctmap[empty_mask] = (0, 0, 0, 0)

    if set_alpha:
        alpha = np.array(sdata.uns['ssam']['max_correlations'], copy=True)
        alpha[alpha < 0] = 0 # drop negative correlations
        alpha = min_r + alpha / (np.max(alpha) / (1.0 - min_r))
        sctmap[..., 3] = alpha

    if rotate == 1 or rotate == 3:
        sctmap = sctmap.swapaxes(0, 1)

    plt.gca().set_facecolor(background)
    plt.imshow(sctmap)
    
    if rotate == 1:
        plt.gca().invert_xaxis()
    elif rotate == 2:
        plt.gca().invert_xaxis()
        plt.gca().invert_yaxis()
    elif rotate == 3:
        plt.gca().invert_yaxis()

    return
